# coding: utf-8
import textwrap
import unittest

import pytest
from lxml import etree
from lxml.builder import ElementMaker

from benker.builders.base_builder import BaseBuilder
from benker.common.lxml_qname import QName
from benker.parsers.cals import CalsParser
from benker.parsers.cals.frame_styles import BORDER_NONE
from benker.parsers.cals.frame_styles import BORDER_SOLID
from benker.schemas import CALS_NS
from benker.schemas import CALS_PREFIX


def cals(name):
    return QName(CALS_NS, name)


class TestCalsParser(unittest.TestCase):
    def setUp(self):
        self.builder = BaseBuilder()

    def test_builder_attached(self):
        parser = CalsParser(self.builder)
        assert parser.builder is self.builder

    def test_ns(self):
        parser = CalsParser(self.builder)
        assert parser.cals_ns is None
        parser = CalsParser(self.builder, cals_ns=CALS_NS)
        assert parser.cals_ns == CALS_NS


class StrBuilder(BaseBuilder):
    """
    Create a ``table`` element with the string representation of the table object.
    """

    def generate_table_tree(self, table):
        element = etree.Element("table")
        element.text = str(table)
        return element


def test_transform_tables():
    # fmt: off
    E = ElementMaker()
    root = E.root(
        E.table(
            E.titles("titles"),
            E.tgroup(
                E.colspec(),
                E.colspec(),
                E.thead(E.row(E.entry("Col A"), E.entry("Col B"))),
                E.tfoot(E.row(E.entry("Foot notes", namest="c1", nameend="c2"))),
                E.tbody(E.row(E.entry("A1"), E.entry("B1")),
                        E.row(E.entry("A2"), E.entry("B2"))),
            ),
        )
    )
    parser = CalsParser(StrBuilder())
    parser.transform_tables(root)
    table = root.xpath("//table")[0]
    assert table.text == textwrap.dedent("""\
    +-----------+-----------+
    |   Col A   |   Col B   |
    +-----------------------+
    | Foot note             |
    +-----------+-----------+
    |    A1     |    B1     |
    +-----------+-----------+
    |    A2     |    B2     |
    +-----------+-----------+""")
    # fmt: on


def test_parse_table():
    # fmt: off
    E = ElementMaker()
    cals_table = E.table(
        E.titles("titles"),
        E.tgroup(
            E.colspec(),
            E.colspec(),
            E.thead(E.row(E.entry("Col A"), E.entry("Col B"))),
            E.tfoot(E.row(E.entry("Foot notes", namest="c1", nameend="c2"))),
            E.tbody(E.row(E.entry("A1"), E.entry("B1")),
                    E.row(E.entry("A2"), E.entry("B2"))),
        ),
    )
    parser = CalsParser(BaseBuilder())
    table = parser.parse_table(cals_table)
    assert len(table.cols) == 2
    assert len(table.rows) == 4
    assert str(table) == textwrap.dedent("""\
    +-----------+-----------+
    |   Col A   |   Col B   |
    +-----------------------+
    | Foot note             |
    +-----------+-----------+
    |    A1     |    B1     |
    +-----------+-----------+
    |    A2     |    B2     |
    +-----------+-----------+""")
    # fmt: on


def test_setup_table():
    parser = CalsParser(BaseBuilder())
    state = parser.setup_table({"key": "value"}, nature="something")
    assert state.table is not None
    assert state.table.styles == {"key": "value"}
    assert state.table.nature == "something"


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        (
            {"frame": "all"},
            {
                "border-bottom": BORDER_SOLID,
                "border-left": BORDER_SOLID,
                "border-right": BORDER_SOLID,
                "border-top": BORDER_SOLID,
            },
            None,
        ),
        ({"colsep": "0"}, {"x-cell-border-right": BORDER_NONE}, None),
        ({"colsep": "1"}, {"x-cell-border-right": BORDER_SOLID}, None),
        ({"rowsep": "0"}, {"x-cell-border-bottom": BORDER_NONE}, None),
        ({"rowsep": "1"}, {"x-cell-border-bottom": BORDER_SOLID}, None),
        ({"orient": "land"}, {"x-sect-orient": "landscape"}, None),
        ({"orient": "port"}, {"x-sect-orient": "portrait"}, None),
        ({"pgwide": "0"}, {"x-sect-cols": "2"}, None),
        ({"pgwide": "1"}, {"x-sect-cols": "1"}, None),
        ({"bgcolor": "#00ff00"}, {"background-color": "#00ff00"}, None),
        ({"tabstyle": "BeautifulTable"}, {}, "BeautifulTable"),
        ({"width": "162"}, {'width': '162.00mm'}, None),
    ],
)
def test_parse_cals_table(attrib, styles, nature):
    # --without namespaces
    cals_table = etree.Element("table", attrib=attrib)
    parser = CalsParser(BaseBuilder())
    state = parser.parse_cals_table(cals_table)
    table = state.table
    assert table.styles == styles
    assert table.nature == nature

    # --with default namespaces
    cals_table_attrib = {cals(k): v for k, v in attrib.items()}
    cals_table = etree.Element(cals("table"), attrib=cals_table_attrib, nsmap={None: CALS_NS})
    parser = CalsParser(BaseBuilder(), cals_ns=CALS_NS)
    state = parser.parse_cals_table(cals_table)
    table = state.table
    assert table.styles == styles
    assert table.nature == nature

    # --with namespaces prefix
    cals_table_attrib = {cals(k): v for k, v in attrib.items()}
    cals_table = etree.Element(cals("table"), attrib=cals_table_attrib, nsmap={CALS_PREFIX: CALS_NS})
    parser = CalsParser(BaseBuilder(), cals_ns=CALS_NS)
    state = parser.parse_cals_table(cals_table)
    table = state.table
    assert table.styles == styles
    assert table.nature == nature


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({"colsep": "0"}, {"x-cell-border-right": BORDER_NONE}, None),
        ({"colsep": "1"}, {"x-cell-border-right": BORDER_SOLID}, None),
        ({"rowsep": "0"}, {"x-cell-border-bottom": BORDER_NONE}, None),
        ({"rowsep": "1"}, {"x-cell-border-bottom": BORDER_SOLID}, None),
        ({"tgroupstyle": "BeautifulTable"}, {}, "BeautifulTable"),
    ],
)
def test_parse_cals_tgroup(attrib, styles, nature):
    # --without namespaces
    cals_tgroup = etree.Element("tgroup", attrib=attrib)
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    state = parser.parse_cals_tgroup(cals_tgroup)
    table = state.table
    assert table.styles == styles
    assert table.nature == nature

    # --with default namespaces
    cals_tgroup_attrib = {cals(k): v for k, v in attrib.items()}
    cals_tgroup = etree.Element(cals("tgroup"), attrib=cals_tgroup_attrib, nsmap={None: CALS_NS})
    parser = CalsParser(BaseBuilder(), cals_ns=CALS_NS)
    parser.setup_table()
    state = parser.parse_cals_tgroup(cals_tgroup)
    table = state.table
    assert table.styles == styles
    assert table.nature == nature


def test_parse_cals_tgroup__overrides_table():
    # --without namespaces
    parser = CalsParser(BaseBuilder())
    parser.setup_table(
        {
            "x-cell-border-right": BORDER_NONE,
            "x-cell-border-bottom": BORDER_NONE,
            "x-sect-orient": "landscape",
            "x-sect-cols": "1",
            "background-color": "velvet",
        },
        "TableOld",
    )
    cals_tgroup_attrib = {"colsep": "1", "rowsep": "1", "tgroupstyle": "TableNew"}
    cals_tgroup = etree.Element("tgroup", attrib=cals_tgroup_attrib)
    state = parser.parse_cals_tgroup(cals_tgroup)
    table = state.table
    assert table.styles == {
        "x-cell-border-right": BORDER_SOLID,
        "x-cell-border-bottom": BORDER_SOLID,
        "x-sect-orient": "landscape",  # preserved
        "x-sect-cols": "1",  # preserved
        "background-color": "velvet",  # preserved
    }
    assert table.nature == "TableNew"


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({"valign": "top"}, {"vertical-align": "top"}, None),
        ({"valign": "middle"}, {"vertical-align": "middle"}, None),
        ({"valign": "bottom"}, {"vertical-align": "bottom"}, None),
        ({"rowsep": "0"}, {"border-bottom": BORDER_NONE}, None),
        ({"rowsep": "1"}, {"border-bottom": BORDER_SOLID}, None),
        ({"rowstyle": "RowStyle"}, {"rowstyle": "RowStyle"}, None),
    ],
)
def test_parse_cals_row(attrib, styles, nature):
    # --without namespaces
    cals_row = etree.Element("row", attrib=attrib)
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    parser._state.next_row()
    state = parser.parse_cals_row(cals_row)
    row = state.table.rows[1]
    assert row.styles == styles
    assert row.nature == nature

    # --with default namespaces
    cals_row_attrib = {cals(k): v for k, v in attrib.items()}
    cals_row = etree.Element(cals("row"), attrib=cals_row_attrib, nsmap={None: CALS_NS})
    parser = CalsParser(BaseBuilder(), cals_ns=CALS_NS)
    parser.setup_table()
    parser._state.next_row()
    state = parser.parse_cals_row(cals_row)
    row = state.table.rows[1]
    assert row.styles == styles
    assert row.nature == nature


@pytest.mark.parametrize("tag, nature", [("thead", "header"), ("tfoot", "footer"), ("tbody", "body")])
def test_parse_cals_row__nature_from_parent(tag, nature):
    # --without namespaces
    cals_parent = etree.Element(tag)
    cals_row = etree.SubElement(cals_parent, "row")
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    parser._state.next_row()
    state = parser.parse_cals_row(cals_row)
    row = state.table.rows[1]
    assert row.nature == nature


def test_parse_cals_row__overrides_parent_valign():
    # --without namespaces
    cals_tbody = etree.Element("tbody", valign="top")
    cals_row = etree.SubElement(cals_tbody, "row", valign="middle")
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    parser._state.next_row()
    state = parser.parse_cals_row(cals_row)
    row = state.table.rows[1]
    assert row.styles["vertical-align"] == "middle"


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({"colsep": "0"}, {"border-right": BORDER_NONE}, None),
        ({"colsep": "1"}, {"border-right": BORDER_SOLID}, None),
        ({"rowsep": "0"}, {"border-bottom": BORDER_NONE}, None),
        ({"rowsep": "1"}, {"border-bottom": BORDER_SOLID}, None),
        ({"bgcolor": "purple"}, {"background-color": "purple"}, None),
        ({"namest": "1"}, {}, None),
        ({"nameend": "2"}, {}, None),
        ({"namest": "1", "nameend": "2"}, {}, None),
        ({"morerows": "2"}, {}, None),
        ({"valign": "top"}, {"vertical-align": "top"}, None),
        ({"valign": "middle"}, {"vertical-align": "middle"}, None),
        ({"valign": "bottom"}, {"vertical-align": "bottom"}, None),
        ({"align": "left"}, {"align": "left"}, None),
        ({"align": "right"}, {"align": "right"}, None),
        ({"align": "center"}, {"align": "center"}, None),
        ({"align": "justify"}, {"align": "justify"}, None),
        ({"align": "char"}, {"align": "left"}, None),
    ],
)
def test_parse_cals_entry(attrib, styles, nature):
    # --without namespaces
    cals_entry = etree.Element("entry", attrib=attrib)
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    state = parser._state
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state.next_col()
    state = parser.parse_cals_entry(cals_entry)
    cell = state.table[(1, 1)]
    # Ignore cell styles extensions (like 'x-cell-empty').
    actual = {k: v for k, v in cell.styles.items() if not k.startswith("x-cell-")}
    assert actual == styles
    assert cell.nature == nature

    # --with default namespaces
    cals_entry_attrib = {cals(k): v for k, v in attrib.items()}
    cals_entry = etree.Element(cals("entry"), attrib=cals_entry_attrib, nsmap={None: CALS_NS})
    parser = CalsParser(BaseBuilder(), cals_ns=CALS_NS)
    parser.setup_table()
    state = parser._state
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state.next_col()
    state = parser.parse_cals_entry(cals_entry)
    cell = state.table[(1, 1)]
    # Ignore cell styles extensions (like 'x-cell-empty').
    actual = {k: v for k, v in cell.styles.items() if not k.startswith("x-cell-")}
    assert actual == styles
    assert cell.nature == nature


@pytest.mark.parametrize(
    "attrib, size",
    [
        ({"namest": "1"}, (1, 1)),
        ({"nameend": "2"}, (2, 1)),
        ({"namest": "1", "nameend": "3"}, (3, 1)),
        ({"morerows": "2"}, (1, 3)),
        ({"namest": "3", "nameend": "5", "morerows": "1"}, (3, 2)),
    ],
)
def test_parse_cals_entry__spanning(attrib, size):
    # --without namespaces
    cals_entry = etree.Element("entry", attrib=attrib)
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    state = parser._state
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state.next_col()
    state = parser.parse_cals_entry(cals_entry)
    cell = state.table[(1, 1)]
    assert cell.size == size


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({"colnum": "2"}, {}, None),
        ({"colsep": "0"}, {"border-right": BORDER_NONE}, None),
        ({"colsep": "1"}, {"border-right": BORDER_SOLID}, None),
        ({"rowsep": "0"}, {"border-bottom": BORDER_NONE}, None),
        ({"rowsep": "1"}, {"border-bottom": BORDER_SOLID}, None),
        ({"colwidth": "5mm"}, {"width": "5mm"}, None),
        ({"align": "left"}, {"align": "left"}, None),
        ({"align": "right"}, {"align": "right"}, None),
        ({"align": "center"}, {"align": "center"}, None),
        ({"align": "justify"}, {"align": "justify"}, None),
        ({"align": "char"}, {"align": "left"}, None),
    ],
)
def test_parse_cals_colspec(attrib, styles, nature):
    # --without namespaces
    cals_colspec = etree.Element("colspec", attrib=attrib)
    parser = CalsParser(BaseBuilder())
    parser.setup_table()
    state = parser._state
    state.next_col()
    state = parser.parse_cals_colspec(cals_colspec)
    col_pos = int(attrib.get("colnum", 1))
    col = state.table.cols[col_pos]
    assert col.styles == styles
    assert col.nature == nature

    # --with default namespaces
    cals_colspec_attrib = {cals(k): v for k, v in attrib.items()}
    cals_colspec = etree.Element(cals("colspec"), attrib=cals_colspec_attrib, nsmap={None: CALS_NS})
    parser = CalsParser(BaseBuilder(), cals_ns=CALS_NS)
    parser.setup_table()
    state = parser._state
    state.next_col()
    state = parser.parse_cals_colspec(cals_colspec)
    col_pos = int(attrib.get("colnum", 1))
    col = state.table.cols[col_pos]
    assert col.styles == styles
    assert col.nature == nature



# coding: utf-8
"""
Grid Cell
=========

A :class:`~benker.cell.Cell` object stores the *content* of a :class:`~benker.grid.Grid` cell.

A cell can have *styles*, a dictionary of key-value properties attached to the cell.

A cell has a *nature* to distinguish between header, body and footer cells.
The default *nature* is ``None``, but you can also use "body", "header", "footer" or whatever…

A cell has top-left coordinates: *x* and *y*. The default coordinates is (1, 1):
this is the top-left coordinate of the cell box.
The coordinates *x* and *y* cannot be null: grid coordinates are 1-indexed.

A cell has a size: *width* and *height*. The default size is (1, 1), you can
increase them to represent horizontal or vertical spanning.
The *width* and the *height* cannot be null.

.. doctest:: cell_demo
    :hide:

    >>> from benker.coord import Coord
    >>> from benker.box import Box
    >>> from benker.cell import Cell

To instantiate a :class:`~benker.cell.Cell`, you can do:

.. doctest:: cell_demo

    >>> c1 = Cell("c1")
    >>> c2 = Cell("c2", styles={'color': 'red'})
    >>> c3 = Cell("c3", nature="footer")
    >>> c4 = Cell("c4", width=2)
    >>> c5 = Cell("c5", height=2)

The string representation of a cell is the string representation of it's content:

.. doctest:: cell_demo

    >>> for cell in c1, c2, c3, c4, c5:
    ...     print(cell)
    c1
    c2
    c3
    c4
    c5

On initialization, the cell min position is always (1, 1), a.k.a. the top-left.

.. doctest:: cell_demo

    >>> c1 = Cell("c1")
    >>> c1.min
    Coord(x=1, y=1)
    >>> c1.size
    Size(width=1, height=1)
    >>> c1.box
    Box(min=Coord(x=1, y=1), max=Coord(x=1, y=1))

A cell can be moved to another position:

.. doctest:: cell_demo

    >>> c1 = Cell("c1", width=3, height=2)
    >>> c2 = c1.move_to(Coord(5, 3))
    >>> c2.min
    Coord(x=5, y=3)
    >>> c2.size
    Size(width=3, height=2)
    >>> c2.box
    Box(min=Coord(x=5, y=3), max=Coord(x=7, y=4))

You can check if a coord is inside the box:

.. doctest:: cell_demo

    >>> c1 = Cell("c1", width=3, height=2)
    >>> c2 = c1.move_to(Coord(5, 3))
    >>> (7, 4) in c2
    True
    >>> Coord(6, 3) in c2
    True
    >>> Box(6, 3, 7, 4) in c2
    True

"""
import functools
import numbers
import sys

from benker.box import Box
from benker.styled import Styled

try:
    # noinspection PyCompatibility
    from collections.abc import Sequence
except ImportError:
    # noinspection PyCompatibility
    from collections import Sequence

PY2 = sys.version_info[0] == 2

text_type = type(u"")
binary_type = type(b"")
string_types = text_type, binary_type


def _safe_text(t):
    if t is None:
        return u""
    elif isinstance(t, binary_type):
        return t.decode('utf-8')
    else:
        return t


def _get_node_text(node):
    if isinstance(node.tag, string_types):
        # etree._Element
        return _safe_text(node.text) + u"".join(map(_get_node_text, node)) + _safe_text(node.tail)
    else:
        # etree._ProcessingInstruction or etree._Comment
        return _safe_text(node.tail)


def get_content_text(content):
    """
    Try hard to extract a good string representation of the cell content.

    >>> from benker.cell import get_content_text

    >>> get_content_text(None) == ''
    True
    >>> get_content_text('') == ''
    True
    >>> print(get_content_text('Hi'))
    Hi
    >>> print(get_content_text(True))
    True
    >>> print(get_content_text(123))
    123
    >>> print(get_content_text(3.14))
    3.14
    >>> get_content_text([None, None]) == ''
    True
    >>> print(get_content_text(["hello", " ", "world!"]))
    hello world!

    :param content: Cell content.

    :return: the cell text:

        - if the content is ``None``: returns "",
        - if the content is a string: return the string unchanged,
        - if the content is a number: return the string representation of the number,
        - if the content is a list of strings, return the concatenated strings (``None`` is ignored),
        - if the content is a list of XML nodes, return the concatenated strings of the elements
          (the processing-instruction and the comments are ignored),
        - else: return a concatenation of the string representation of the content items.

    .. versionadded:: 0.4.1

    .. versionchanged:: 0.5.1
       Returns the XML Element text and tail text.
    """
    if content is None:
        return u""
    if isinstance(content, text_type):
        return content
    if isinstance(content, binary_type):
        return content.decode('utf-8')  # PY2
    if isinstance(content, numbers.Number):
        return text_type(content)
    if isinstance(content, Sequence):
        return u"".join(get_content_text(node) for node in content)
    if hasattr(content, 'tag'):
        return _get_node_text(content)
    return text_type(content)


@functools.total_ordering
class Cell(Styled):
    """
    Cell of a grid.

    :ivar content: user-defined cell content. It can be of any type: ``None``,
        :class:`str`, :class:`int`, :class:`float`, a container (:class:`list`),
        a XML element, etc. The same content can be shared by several cells, it's
        your own responsibility to handle the copy (or deep copy) of the *content*
        reference when needed.

        .. note::

            In a :class:`~benker.grid.Grid`, the :ref:`merging <benker__grid__merging>`
            of two cell contents is done with the "+" operator (:func:`~operator.__add__`).
            You can override this by using a *content_appender*, a two-arguments
            function which will perform the concatenation of the two contents.

    .. versionchanged:: 0.4.2
       The default value of *nature* is ``None`` (instead of "body").
    """

    __slots__ = ('content', '_box')

    def __init__(self, content, styles=None, nature=None, x=1, y=1, width=1, height=1):
        super(Cell, self).__init__(styles, nature)
        self.content = content
        self._box = Box(x, y, x + width - 1, y + height - 1)

    def __str__(self):
        """
        Try hard to extract a good string representation of the cell.

        :return: the cell text:

            - if the content is ``None``: returns "",
            - if the content is a string: return the string unchanged,
            - if the content is a number: return the string representation of the number,
            - if the content is a list of strings, return the concatenated strings (``None`` is ignored),
            - if the content is a list of XML nodes, return the concatenated strings of the elements
              (the processing-instruction and the comments are ignored),
            - else: return a concatenation of the string representation of the content items.

        .. versionchanged:: 0.4.1
           Improve the string representation of a cell in order to extract the cell text
           even if it contains a list of strings or XML nodes.
        """
        return get_content_text(self.content)

    if PY2:
        __unicode__ = __str__

        def __str__(self):
            return self.__unicode__().encode('ascii', errors='backslashreplace')

    def __repr__(self):
        cls = self.__class__.__name__
        return (
            "<{cls}({self.content!r}, "
            "styles={self.styles!r}, "
            "nature={self.nature!r}, "
            "x={self.min.x!r}, "
            "y={self.min.y!r}, "
            "width={self.width!r}, "
            "height={self.height!r})>".format(cls=cls, self=self)
        )

    @property
    def box(self):
        """ Bounding box of the cell. """
        return self._box

    @property
    def min(self):
        """ Minimum coordinates of the cell -- top-left coordinates. """
        return self.box.min

    @property
    def max(self):
        """ Maximum coordinates of the cell -- bottom-right coordinates. """
        return self.box.max

    @property
    def size(self):
        """ Size of the cell -- (*with*, *height*). """
        return self.box.size

    @property
    def width(self):
        """ Width of the cell -- columns spanning. """
        return self.box.width

    @property
    def height(self):
        """ Height of the cell -- rows spanning. """
        return self.box.height

    def __contains__(self, other):
        return other in self.box

    # total ordering based on Box ordering

    def __lt__(self, other):
        if isinstance(other, Cell):
            return self.box < other.box
        elif isinstance(other, Box):
            return self.box < other
        return NotImplemented

    def transform(self, coord=None, size=None):
        """
        Transform the bounding box of the cell by making a move and a resize.

        :type  coord: tuple[int, int] or benker.coord.Coord
        :param coord: new top-left coordinates of the cell, by default it is unchanged.

        :type  size: tuple[int, int] or benker.size.Size
        :param size: new size of the cell, by default it is unchanged.

        :rtype: benker.cell.Cell
        :return: a copy of this cell after transformation.
        """
        box = self.box.transform(coord=coord, size=size)
        cell = Cell(
            self.content,
            styles=self.styles,
            nature=self.nature,
            x=box.min.x,
            y=box.min.y,
            width=box.width,
            height=box.height,
        )
        return cell

    def move_to(self, coord):
        """
        Move the cell to the given coordinates.

        See: :meth:`~benker.cell.Cell.transform`.

        :type  coord: tuple[int, int] or benker.coord.Coord
        :param coord: new top-left coordinates of the cell, by default it is unchanged.

        :rtype: benker.cell.Cell
        :return: a copy of this cell after transformation.
        """
        return self.transform(coord=coord)

    def resize(self, size):
        """
        Resize the cell to the given size.

        See: :meth:`~benker.cell.Cell.transform`.

        :type  size: tuple[int, int] or benker.size.Size
        :param size: new size of the cell, by default it is unchanged.

        :rtype: benker.cell.Cell
        :return: a copy of this cell after transformation.
        """
        return self.transform(size=size)



# coding: utf-8
from __future__ import print_function

import shutil

import py.path  # type hints
import pytest
import xmldiff.main
from lxml import etree

from benker.converters.cals2formex import convert_cals2formex

from tests.resources import RESOURCES_DIR


@pytest.mark.parametrize(
    "input_name, expected_name, cals_prefix, cals_ns",
    # fmt: off
    [
        (
            "cals/tbl_small_table.xml",
            "cals2formex/tbl_small_table.xml",
            "cals",
            "https://lib.benker.com/schemas/cals.xsd",
        ),
        (
            "cals/tbl_sample.xml",
            "cals2formex/tbl_sample.xml",
            "cals",
            "https://lib.benker.com/schemas/cals.xsd",
        ),
        (
            "cals/tbl_sample_formex.xml",
            "cals2formex/tbl_sample_formex.xml",
            "cals",
            "https://lib.benker.com/schemas/cals.xsd",
        ),
        (
            "cals/tbl_sample_cals2.xml",
            "cals2formex/tbl_sample_cals2.xml",
            None,
            None,
        ),
        (
            "cals/fix_9_lost_text_in_entries.xml",
            "cals2formex/fix_9_lost_text_in_entries.xml",
            None,
            None,
        ),
        (
            "cals/fix_11_corpus_width_missing.xml",
            "cals2formex/fix_11_corpus_width_missing.xml",
            None,
            None,
        ),
        (
            "cals/fix_12_missing_colspec_attributes.xml",
            "cals2formex/fix_12_missing_colspec_attributes.xml",
            None,
            None,
        ),
        (
            "cals/fix_10_embed_gr_notes1.xml",
            "cals2formex/fix_10_embed_gr_notes1.xml",
            None,
            None,
        ),
        (
            "cals/fix_10_embed_gr_notes2.xml",
            "cals2formex/fix_10_embed_gr_notes2.xml",
            None,
            None,
        ),
        (
            "cals/tbl_col_type_header.xml",
            "cals2formex/tbl_col_type_header.xml",
            None,
            None,
        ),
    ],
    # fmt: on
)
def test_convert_cals2formex(input_name, expected_name, cals_prefix, cals_ns, tmpdir):
    # type: (str, str, str, str, py.path.local) -> None
    src_xml = RESOURCES_DIR.join(input_name)  # type: py.path.local
    dst_xml = tmpdir.join(src_xml.basename)
    convert_cals2formex(
        str(src_xml), str(dst_xml), width_unit="mm", use_cals=True, cals_prefix=cals_prefix, cals_ns=cals_ns
    )

    # - Compare with expected
    xml_parser = etree.XMLParser(remove_blank_text=True)
    expected_xml = RESOURCES_DIR.join(expected_name)  # type: py.path.local
    if expected_xml.exists():
        expected_tree = etree.parse(str(expected_xml), parser=xml_parser)
        expected_elements = expected_tree.xpath("//TBL")
        dst_tree = etree.parse(str(dst_xml), parser=xml_parser)
        dst_elements = dst_tree.xpath("//TBL")
        assert len(expected_elements) == len(dst_elements)

        for dst_elem, expected_elem in zip(dst_elements, expected_elements):
            diff_list = xmldiff.main.diff_trees(dst_elem, expected_elem)
            assert not diff_list
    else:
        # missing resource: create it
        shutil.copy(str(dst_xml), str(expected_xml))
        print("You should check and commit the file: '{}'".format(expected_xml))



# coding: utf-8
import unittest

import pytest
from lxml import etree

from benker.builders.base_builder import BaseBuilder


class TestBaseBuilder(unittest.TestCase):
    def test_init(self):
        builder = BaseBuilder()
        assert builder.options == {}

    def test_init__options(self):
        builder = BaseBuilder(key="value")
        assert builder.options == {"key": "value"}


def pi(target, text=None, tail=None):
    node = etree.ProcessingInstruction(target, text=text)
    node.tail = tail
    return node


@pytest.mark.parametrize(
    u"elements, expected",
    # fmt: off
    [
        (
            None,
            u"<cell/>",
        ),
        (
            u"text",
            u"<cell>text</cell>",
        ),
        (
            b"text",
            u"<cell>text</cell>",
        ),
        (
            3.14,
            u"<cell>3.14</cell>",
        ),
        (
            etree.Element(u"P"),
            u"<cell><P/></cell>",
        ),
        (
            [etree.Element(u"P")],
            u"<cell><P/></cell>",
        ),
        (
            [u"text"],
            u"<cell>text</cell>",
        ),
        (
            [u"text", pi(u"instr", text=u"no='1'")],
            u"<cell>text<?instr no='1'?></cell>",
        ),
        (
            [pi(u"instr", text=u"no='1'")],
            u"<cell><?instr no='1'?></cell>",
        ),
        (
            [pi(u"instr", text=u"no='1'", tail=u" following")],
            u"<cell><?instr no='1'?> following</cell>",
        ),
        (
            [u"one", pi(u"instr", text=u"no='1'", tail=u"two")],
            u"<cell>one<?instr no='1'?>two</cell>",
        ),
        (
            [u"one", pi(u"instr", text=u"no='1'", tail=u"two"), u" three", pi(u"instr", text=u"no='2'", tail=u"four")],
            u"<cell>one<?instr no='1'?>two three<?instr no='2'?>four</cell>",
        ),
        (
            [u"one", etree.XML(u"<b>bold</b>"), u"two"],
            u"<cell>one<b>bold</b>two</cell>",
        ),
    ],
    # fmt: on
)
def test_append_content(elements, expected):
    builder = BaseBuilder()
    cell_elem = etree.Element(u"cell")
    builder.append_cell_elements(cell_elem, elements)
    assert etree.tounicode(cell_elem) == expected



# coding: utf-8
"""
.. _available-parsers:

Available Parsers
=================

This package contains a collection of parsers which you can use in conjonction
with builders to convert tables from one format to another.

You can pick a builder in the :ref:`Available Builders <available-builders>`.
"""



# coding: utf-8
from __future__ import print_function

import sys
import unittest

import pytest
import xmldiff.main
from lxml import etree

from benker.builders.cals import CalsBuilder
from benker.cell import Cell
from benker.table import Table

# noinspection PyProtectedMember
ElementType = etree._Element


class TestCalsBuilder(unittest.TestCase):
    def test_init(self):
        builder = CalsBuilder()
        assert builder.cals_ns.uri is None
        assert builder.cals_ns.prefix is None
        assert builder.ns_map == {}
        assert builder.width_unit == "mm"
        assert builder.table_in_tgroup is False
        assert builder.tgroup_sorting == {"header": 0, "footer": 1, "body": 2}

    def test_option_cals_ns(self):
        builder = CalsBuilder(cals_ns="http://cals")
        assert builder.cals_ns.uri == "http://cals"
        assert builder.ns_map == {None: "http://cals"}

    def test_option_cals_prefix(self):
        builder = CalsBuilder(cals_ns="http://cals", cals_prefix="cals")
        assert builder.cals_ns.prefix == "cals"
        assert builder.ns_map == {"cals": "http://cals"}

    def test_option_cals_prefix__without_ns(self):
        with self.assertRaises(ValueError):
            CalsBuilder(cals_prefix="cals")

    def test_option_width_unit(self):
        builder = CalsBuilder(width_unit="pt")
        assert builder.width_unit == "pt"

    def test_option_table_in_tgroup(self):
        builder = CalsBuilder(table_in_tgroup=True)
        assert builder.table_in_tgroup is True

    def test_option_tgroup_sorting(self):
        builder = CalsBuilder(tgroup_sorting=["header", "body", "footer"])
        assert builder.tgroup_sorting == {"header": 0, "body": 1, "footer": 2}

    def test_option_tgroup_sorting__missing(self):
        with pytest.raises(ValueError):
            CalsBuilder(tgroup_sorting=["header", "body"])


def test_setup_table():
    builder = CalsBuilder()
    table = Table()
    result = builder.setup_table(table)
    assert result == table


TEST_DATA__WITH_SEP = [
    # fmt: off
    pytest.param(
        {},
        {},
        id="no_style"),

    pytest.param(
        {
            'border-top': 'double 1.0pt #FFFFFF',
            'border-right': 'dotted 2.0pt #FF0000',
            'border-bottom': 'dashed 3.0pt #00FF00',
            'border-left': 'outset 4.0pt #0000FF',
        },
        {'colsep': '1', 'rowsep': '1'},
        id="style_border-all"),

    pytest.param(
        {'border-top': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-top"),

    pytest.param(
        {'border-right': 'double 1.0pt #FFFFFF'},
        {'colsep': '1'},
        id="style_border-right"),

    pytest.param(
        {'border-bottom': 'double 1.0pt #FFFFFF'},
        {'rowsep': '1'},
        id="style_border-bottom"),

    pytest.param(
        {'border-left': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-left"),
    # fmt: on
]


@pytest.mark.parametrize("cell_styles, expected", TEST_DATA__WITH_SEP)
def test_build_cell__with_sep(cell_styles, expected):
    # .. note::
    #
    #   CALS default value for @colsep and @rowsep is "1".

    builder = CalsBuilder()

    # -- create a minimal <row> element
    row_elem = etree.XML("<row/>")

    # -- setup a minimal cell
    p_elem = etree.XML("<p>text</p>")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, styles=cell_styles)
    cell_x2_y1 = Cell([p_elem], x=2, y=1, styles=cell_styles)
    cell_x1_y2 = Cell([p_elem], x=1, y=2, styles=cell_styles)
    cell_x2_y2 = Cell([p_elem], x=2, y=2, styles=cell_styles)
    builder.setup_table(Table([cell_x1_y1, cell_x2_y1, cell_x1_y2, cell_x2_y2]))

    # -- build the cell
    builder.build_cell(row_elem, cell_x1_y1)

    # -- check the '<entry>' attributes
    entry_elem = row_elem[0]  # type: ElementType
    assert entry_elem.tag == u"entry"
    assert entry_elem.attrib == expected
    assert entry_elem[0] == p_elem


TEST_DATA__WITHOUT_SEP = [
    # fmt: off
    pytest.param(
        {},
        {},
        id="no_style"),

    pytest.param(
        {
            'border-top': 'double 1.0pt #FFFFFF',
            'border-right': 'dotted 2.0pt #FF0000',
            'border-bottom': 'dashed 3.0pt #00FF00',
            'border-left': 'outset 4.0pt #0000FF',
        },
        {'colsep': '1', 'rowsep': '1'},
        id="style_border-all"),

    pytest.param(
        {'border-top': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-top"),

    pytest.param(
        {'border-right': 'double 1.0pt #FFFFFF'},
        {'colsep': '1'},
        id="style_border-right"),

    pytest.param(
        {'border-bottom': 'double 1.0pt #FFFFFF'},
        {'rowsep': '1'},
        id="style_border-bottom"),

    pytest.param(
        {'border-left': 'double 1.0pt #FFFFFF'},
        {},
        id="style_border-left"),
    # fmt: on
]


@pytest.mark.parametrize("cell_styles, expected", TEST_DATA__WITHOUT_SEP)
def test_build_cell__without_sep(cell_styles, expected):
    # .. note::
    #
    #   CALS default value for @colsep and @rowsep is "1"
    #   For testing, we set it to "0" by default.

    # -- create a minimal <row> element
    row_elem = etree.XML("<row/>")

    # -- setup a minimal cell
    p_elem = etree.XML("<p>text</p>")
    cell_x1_y1 = Cell([p_elem], x=1, y=1, styles=cell_styles)
    cell_x2_y1 = Cell([p_elem], x=2, y=1, styles=cell_styles)
    cell_x1_y2 = Cell([p_elem], x=1, y=2, styles=cell_styles)
    cell_x2_y2 = Cell([p_elem], x=2, y=2, styles=cell_styles)

    # -- build the cell
    builder = CalsBuilder()
    builder.setup_table(Table([cell_x1_y1, cell_x2_y1, cell_x1_y2, cell_x2_y2]))
    builder.build_cell(row_elem, cell_x1_y1)

    # -- check the '<entry>' attributes
    entry_elem = row_elem[0]  # type: ElementType
    assert entry_elem.tag == u"entry"
    assert entry_elem.attrib == expected
    assert entry_elem[0] == p_elem


TEST_DATA__ALIGN = [
    pytest.param(
        {"vertical-align": "top", "align": "left"},
        {"valign": "top", "align": "left"},
        id="style_vertical-align_top_align_left",
    ),
    pytest.param(
        {"vertical-align": "bottom", "align": "right"},
        {"valign": "bottom", "align": "right"},
        id="style_vertical-align_bottom_align_right",
    ),
    pytest.param(
        {"vertical-align": "middle", "align": "center"},
        {"valign": "middle", "align": "center"},
        id="style_vertical-align_middle_align_center",
    ),
    pytest.param(
        {"vertical-align": "w-both", "align": "justify"},
        {"valign": "bottom", "align": "justify"},
        id="style_vertical-align_both_align_justify",
    ),
]


@pytest.mark.parametrize("cell_styles, expected", TEST_DATA__ALIGN)
def test_build_cell__align(cell_styles, expected):
    builder = CalsBuilder()

    # -- create a minimal <row> element
    row_elem = etree.XML("<row/>")

    # -- setup a minimal cell
    p_elem = etree.XML("<p>text</p>")
    cell = Cell([p_elem], x=1, y=1, styles=cell_styles)
    builder.setup_table(Table([cell]))

    # -- build the cell
    builder.build_cell(row_elem, cell)

    # -- check the '<entry>' attributes
    entry_elem = row_elem[0]  # type: ElementType
    assert entry_elem.tag == u"entry"

    # we don't want to check @colsep/@rowsep here
    actual = dict(entry_elem.attrib)
    actual.pop("colsep", None)
    actual.pop("rowsep", None)

    assert actual == expected
    assert entry_elem[0] == p_elem


@pytest.mark.parametrize(
    "tgroup_sorting, expected_tags",
    [
        (None, ["colspec", "thead", "tfoot", "tbody"]),
        (["header", "body", "footer"], ["colspec", "thead", "tbody", "tfoot"]),
        (["header", "footer", "body"], ["colspec", "thead", "tfoot", "tbody"]),
        (["footer", "body", "header"], ["colspec", "tfoot", "tbody", "thead"]),
    ],
)
def test_build_tgroup__tgroup_sorting(tgroup_sorting, expected_tags):
    # -- create a table with different natures
    table = Table()
    row1 = table.rows[1]
    row2 = table.rows[2]
    row3 = table.rows[3]
    row1.nature = "header"
    row2.nature = "body"
    row3.nature = "footer"
    row1.insert_cell("a")
    row2.insert_cell("b")
    row3.insert_cell("c")

    # -- create a builder
    table_elem = etree.Element("table")
    builder = CalsBuilder(tgroup_sorting=tgroup_sorting)
    builder.setup_table(table)
    builder.build_tgroup(table_elem, table)

    # -- check the tgroup children name and order
    actual_tags = [elem.tag for elem in table_elem.xpath("tgroup/*")]
    assert actual_tags == expected_tags


def test_build_table():
    # see: formex-4/samples/jo-compl-2002C_061/C_2002061EN.01000403.xml

    table = Table(
        styles={
            "border-top": "solid",
            "border-bottom": "solid",
            "x-sect-orient": "landscape",
            "x-sect-cols": "1",
            "background-color": "yellow",
            "width": "247",
        }
    )
    table.rows[1].nature = "header"
    table.rows[1].insert_cell(u"Expert group")
    table.rows[1].insert_cell(u"First name and surname of the expert")
    table.rows[2].insert_cell(u"Control of infectious diseases")
    table.rows[2].insert_cell(u"Michael Angelo BORG")
    table.rows[3].insert_cell(u"Information society", height=3)
    table.rows[3].insert_cell(u"Tony HEY")
    table.rows[4].insert_cell(u"José L. ENCARNAÇÃO")
    table.rows[5].insert_cell(u"Berit SVENDSEN")
    table.rows[6].insert_cell(u"Controlled thermonuclear fusion")
    table.rows[6].insert_cell(u"Pekka PIRILÄ")

    builder = CalsBuilder()
    table_elem = builder.build_table(table)

    xml_parser = etree.XMLParser(remove_blank_text=True)

    # fmt: off
    expected = etree.XML(u"""\
    <table frame="topbot" colsep="0" rowsep="0" orient="land" pgwide="1" bgcolor="yellow" width="247.00mm">
      <tgroup cols="2">
        <colspec colnum="1" colname="c1"/>
        <colspec colnum="2" colname="c2"/>
        <thead>
          <row>
            <entry>Expert group</entry>
            <entry>First name and surname of the expert</entry>
          </row>
        </thead>
        <tbody>
          <row>
            <entry>Control of infectious diseases</entry>
            <entry>Michael Angelo BORG</entry>
          </row>
          <row>
            <entry morerows="2">Information society</entry>
            <entry>Tony HEY</entry>
          </row>
          <row>
            <entry>José L. ENCARNAÇÃO</entry>
          </row>
          <row>
            <entry>Berit SVENDSEN</entry>
          </row>
          <row>
            <entry>Controlled thermonuclear fusion</entry>
            <entry>Pekka PIRILÄ</entry>
          </row>
        </tbody>
      </tgroup>
    </table>""", parser=xml_parser)
    # fmt: on

    for elem in table_elem.xpath("//*"):
        elem.text = elem.text or None
    for elem in expected.xpath("//*"):
        elem.text = elem.text or None

    diff_list = xmldiff.main.diff_trees(table_elem, expected)
    if diff_list:
        print(etree.tounicode(table_elem, pretty_print=True, with_tail=False), file=sys.stderr)
        assert diff_list == []



# coding: utf-8
"""
Units
=====

Utility functions to convert values from one unit to another.
"""
import re

#: Usual units Lengths in meter
UNITS = {
    'mm': 0.001,
    'dm': 0.1,
    'cm': 0.01,
    'm': 1.0,
    'in': 0.001 * 25.4,
    'ft': 0.3048,
    'px': 0.001 * 25.4 / 72,
    'pt': 0.001 * 25.4 / 72,
    'pc': 0.001 * 25.4 / 12,
}


def convert_value(value, unit_in, unit_out):
    """
    Convert a value from one unit to another.

    To convert '1pt' to 'mm', you can do:

    .. doctest:: units_demo

        >>> from benker.units import convert_value

        >>> convert_value(1, 'pt', 'mm')
        0.353

    :type  value: int or float
    :param value: Value to convert

    :param unit_in: Current unit of the value.

    :param unit_out: Expected unit of the value.

    :rtype:  float
    :return: The converted value
    """
    return round(value * UNITS[unit_in] / UNITS[unit_out], 3)


def parse_width(width, default_unit="mm"):
    """
    Parse a width and return the value and its unit.

    .. doctest:: units_demo

        >>> from benker.units import parse_width

        >>> parse_width("210")
        (210.0, 'mm')

        >>> parse_width("210mm")
        (210.0, 'mm')

        >>> parse_width("210pt")
        (210.0, 'pt')

    :param width: width string to parse, for instance: "247mm".
    :param default_unit: default unit to use if it is not specified

    :rtype: (float, str)
    :return: the value and its unit.

    .. versionadded:: 0.5.1
    """
    width, unit = re.findall(r"([+-]?(?:[0-9]*[.])?[0-9]+)(cm|dm|ft|in|mm|pc|pt|px|m)?", width)[0]
    unit = unit or default_unit
    return float(width), unit



# coding: utf-8
"""
Formex 4 Parser
===============

This module can parse the tables (``TBL`` elements) of a Formex 4 file.

The ``TBL`` element is used to mark up a Formex table, which actually contains text structured
in columns with related data.

A table usually contains the following information:

- an optional title (``TITLE``),
- one or more structured text blocks (``GR.SEQ``) in order to mark up optional explanatory
  information about the table content, located between the title of the table and the table itself,
- optionally a group of notes called in the table (``GR.NOTES``),
- the corpus of the table (``CORPUS``).

When building the internal table object, this builder will:

- interpret the title (``TITLE``) and structured text blocks (``GR.SEQ``) like rows.
  The *nature* attribute of each row will be "title" and "text-block" respectively.
- interpret the group of notes (``GR.NOTES``) like a row of *nature* "footer"
- interpret the corpus of the table (``CORPUS``) like the body of the table.
  The *nature* attribute of each row will be "body".

.. note::

   Since the Formex table structure is not suitable for typesetting/page layout, this parser is
   also able to parse CALS-like attributes (for instance ``frame``, ``cols``, ``colsep``,
   ``rowsep``, ...) and CALS-like elements (for instance ``colspec``). This attributes and
   elements may be added with the Formex 4 builder,
   see :class:`~benker.builders.formex.FormexBuilder`.

.. versionadded:: 0.5.0
"""
import re

from lxml import etree

from benker.box import Box
from benker.common.lxml_iterwalk import iterwalk
from benker.common.lxml_qname import QName
from benker.parsers.base_parser import BaseParser
from benker.parsers.cals.frame_styles import BORDER_NONE
from benker.parsers.cals.frame_styles import BORDER_SOLID
from benker.parsers.cals.frame_styles import get_frame_styles
from benker.table import Table

# noinspection PyProtectedMember
#: Element Type
ElementType = etree._Element


class FormexParser(BaseParser):
    """
    Formex 4 tables parser
    """

    def __init__(self, builder, formex_ns=None, cals_ns=None, embed_gr_notes=False, **options):
        """
        Construct a parser

        :type  builder: benker.builders.base_builder.BaseBuilder
        :param builder:
            Builder used by this parser to instantiate :class:`~benker.table.Table` objects.

        :param str formex_ns:
            Namespace to use for Formex elements and attributes.
            Set ``None`` (or "") if you don't use namespace.

        :param str cals_ns:
            Namespace to use for CALS-like elements and attributes.
            Set ``None`` (or "") if you don't use namespace.

        :param bool embed_gr_notes:
            If ``True``, Embed the ``GR.NOTES`` in a row/cell, else only copy the content (not the ``GR.NOTES`` tag).

        :keyword str options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.
        """
        self.formex_ns = formex_ns
        self.cals_ns = cals_ns
        self.embed_gr_notes = embed_gr_notes
        super(FormexParser, self).__init__(builder, **options)

    def get_formex_qname(self, name):
        return QName(self.formex_ns, name)

    def get_cals_qname(self, name):
        return QName(self.cals_ns, name)

    def transform_tables(self, tree):
        if self.formex_ns:
            nodes = tree.xpath("//fmx:CORPUS", namespaces={"fmx": self.formex_ns})
        else:
            nodes = tree.xpath("//CORPUS")
        for node in nodes:
            table = self.parse_table(node)
            table_elem = self.builder.generate_table_tree(table)
            parent = node.getparent()
            index = parent.index(node)
            parent.insert(index, table_elem)
            table_elem.tail = node.tail
            parent.remove(node)

    # noinspection PyPep8Naming
    def parse_table(self, fmx_corpus):
        """
        Convert a ``<CORPUS>`` Formex element into table object.

        :type  fmx_corpus: ElementType
        :param fmx_corpus: Formex element.

        :rtype: ElementType
        :return: Table.
        """
        state = self._state
        state.reset()

        # -- Formex elements
        fmx = self.get_formex_qname

        CORPUS = fmx("CORPUS").text
        ROW = fmx("ROW").text
        CELL = fmx("CELL").text
        MARGIN = fmx("MARGIN").text
        BLK = fmx("BLK").text
        TI_BLK = fmx("TI.BLK").text
        STI_BLK = fmx("STI.BLK").text

        # -- CALS-like elements
        cals = self.get_cals_qname

        colspec = cals("colspec").text

        elements = {CORPUS, ROW, CELL, MARGIN, BLK, TI_BLK, STI_BLK, colspec}
        context = iterwalk(fmx_corpus, events=("start", "end"), tag=elements)

        depth = 0
        for action, elem in context:
            elem_tag = elem.tag
            if elem_tag == CORPUS:
                if action == "start":
                    depth += 1
                else:
                    depth -= 1
            if depth > 1:
                # .. note:: context.skip_subtree() is not available for all version of lxml
                # This <TBL> element is inside the table.
                # It will be handled separately in another call to transform_tables()
                continue
            if action == "start":
                # tags sorted by frequency:
                if elem_tag == CELL:
                    state.next_col()
                    self.parse_fmx_cell(elem)

                elif elem_tag == ROW:
                    state.next_row()
                    self.parse_fmx_row(elem)

                elif elem_tag == BLK:
                    # only a container
                    pass

                elif elem_tag == TI_BLK:
                    state.next_row()
                    self.parse_fmx_ti_blk(elem)

                elif elem_tag == STI_BLK:
                    state.next_row()
                    self.parse_fmx_sti_blk(elem)

                elif elem_tag == colspec:
                    state.next_col()
                    self.parse_fmx_colspec(elem)

                elif elem_tag == CORPUS:
                    self.parse_fmx_corpus(elem)

                elif elem_tag == MARGIN:
                    raise NotImplementedError("MARGIN is not supported yet")

                else:
                    raise NotImplementedError(elem_tag)
            else:
                if elem_tag in {ROW, TI_BLK, STI_BLK}:
                    bounding_box = Box(1, state.row_pos, len(state.table.cols), state.row_pos)
                    state.table.fill_missing(bounding_box, None, nature=state.row.nature)
                elif elem_tag == CORPUS:
                    # if there is a GR.NOTES, we create a row for it with the nature "footer"
                    if self.formex_ns:
                        nodes = elem.xpath("preceding-sibling::fmx:GR.NOTES", namespaces={"fmx": self.formex_ns})
                    else:
                        nodes = elem.xpath("preceding-sibling::GR.NOTES")
                    for fmx_gr_notes in nodes:
                        # Convert the GR.NOTES and remove it
                        state.next_row()
                        self.parse_gr_notes(fmx_gr_notes)
                        fmx_tbl = fmx_gr_notes.getparent()
                        fmx_tbl.remove(fmx_gr_notes)
                    state.table.fill_missing(state.table.bounding_box, None)

        return state.table

    def setup_table(self, styles=None, nature=None):
        table = Table(styles=styles, nature=nature)
        self._state.table = table
        return self._state

    def parse_fmx_corpus(self, fmx_corpus):
        fmx_tbl = fmx_corpus.getparent()
        styles, nature = self.parse_tbl_styles(fmx_tbl)

        # support for CALS-like elements and attributes
        cals = self.get_cals_qname

        # -- attribute @cals:frame
        frame = fmx_corpus.attrib.get(cals("frame"))
        styles.update(get_frame_styles(frame))

        # -- attribute @cals:colsep
        colsep = fmx_corpus.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["x-cell-border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = fmx_corpus.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["x-cell-border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:orient
        orient = fmx_corpus.attrib.get(cals("orient"))
        orient_map = {"land": "landscape", "port": "portrait"}
        if orient in orient_map:
            styles["x-sect-orient"] = orient_map[orient]

        # -- attribute @cals:pgwide
        pgwide = fmx_corpus.attrib.get(cals("pgwide"))
        pgwide_map = {"0": "2", "1": "1"}
        if pgwide in pgwide_map:
            styles["x-sect-cols"] = pgwide_map[pgwide]

        # -- attribute @cals:bgcolor
        bgcolor = fmx_corpus.attrib.get(cals("bgcolor"))
        if bgcolor:
            styles["background-color"] = bgcolor

        # -- attribute @cals:tabstyle
        tabstyle = fmx_corpus.attrib.get(cals("tabstyle"))
        if tabstyle:
            # overrides the calculated tabstyle (see @fmx:CLASS)
            nature = nature + "-" + tabstyle if nature else tabstyle

        return self.setup_table(styles, nature)

    def parse_tbl_styles(self, fmx_tbl):
        """
        Parse a ``TBL`` element and extract the styles

        :type  fmx_tbl: ElementType
        :param fmx_tbl: table

        :return: dictionary of styles and nature
        """
        if fmx_tbl is None:
            # mainly for unit tests
            return {}, None

        fmx = self.get_formex_qname
        styles = {}

        # -- attribute @fmx:NO.SEQ => ignored (computed)

        # -- attribute @fmx:CLASS => @cals:tabstyle or @cals:tgroupstyle
        nature = fmx_tbl.attrib.get(fmx("CLASS"))

        # -- attribute @fmx:COLS => ignored (*table.cols*)

        # -- attribute @fmx:PAGE.SIZE
        page_size = fmx_tbl.attrib.get(fmx("PAGE.SIZE"))
        page_size_map = {
            "DOUBLE.LANDSCAPE": "landscape",
            "DOUBLE.PORTRAIT": "portrait",
            "SINGLE.LANDSCAPE": "landscape",
            "SINGLE.PORTRAIT": "portrait",
        }
        if page_size in page_size_map:
            styles["x-sect-orient"] = page_size_map[page_size]

        return styles, nature

    def parse_fmx_row(self, fmx_row):
        """
        Parse a ``ROW`` element which contains ``CELL`` elements.

        This element may be in a ``BLK```

        :type  fmx_row: ElementType
        :param fmx_row: table row
        """
        fmx = self.get_formex_qname
        styles = {}

        # -- attribute @fmx:TYPE => *nature*
        type_map = {"ALIAS": "header", "HEADER": "header", "NORMAL": "body", "NOTCOL": "body", "TOTAL": "body"}
        row_type = fmx_row.attrib.get(fmx("TYPE"))
        nature = type_map.get(row_type, "body")

        # the @fmx:TYPE is preserved in a @cals:rowstyle
        # the BLK level is also stored in this attribute
        blk_count = self._count_blk(fmx_row)
        blk_level = "level{count}".format(count=blk_count) if blk_count else None
        row_styles = list(filter(None, [row_type, blk_level]))
        if row_styles:
            styles["rowstyle"] = "-".join(["ROW"] + row_styles)

        styles.update(self.parse_cals_row_styles(fmx_row))

        # -- Create a ROW
        state = self._state
        state.row = state.table.rows[state.row_pos]
        state.row.nature = nature
        state.row.styles = styles

        return state  # mainly for unit test

    def parse_cals_row_styles(self, fmx_elem):
        """
        Parse the row styles

        :type  fmx_elem: ElementType
        :param fmx_elem: Formex element: ``ROW``, ``TI.BLK``, ``STI.BLK`` or ``GR.NOTES``.
        :return: CSS-like styles

        .. versionchanged:: 0.5.1
           The "vertical-align" style is built from the ``@cals:valign`` attribute.
        """
        # support for CALS-like elements and attributes
        cals = self.get_cals_qname
        styles = {}

        # -- attribute @cals:valign (extension)
        valign = fmx_elem.attrib.get(cals("valign"))
        valign_map = {'top': 'top', 'middle': 'middle', 'bottom': 'bottom'}
        if valign in valign_map:
            styles["vertical-align"] = valign_map[valign]

        # -- attribute @cals:rowsep
        rowsep = fmx_elem.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:bgcolor
        bgcolor = fmx_elem.attrib.get(cals("bgcolor"))
        if bgcolor:
            styles["background-color"] = bgcolor

        # -- attribute @cals:rowstyle (extension)
        rowstyle = fmx_elem.attrib.get(cals("rowstyle"))
        if rowstyle:
            # overrides the previously calculated @cals:rowstyle attribute
            styles["rowstyle"] = rowstyle

        return styles

    def parse_fmx_ti_blk(self, fmx_ti_blk):
        """
        Parse a ``TI.BLK`` element, considering it like a row of a single cell.

        For instance:

        .. code-block:: xml

           <TI.BLK COL.START="1" COL.END="2">
             <P><HT TYPE="BOLD">TI.BLK title</HT></P>
           </TI.BLK>

        :type  fmx_ti_blk: ElementType
        :param fmx_ti_blk: title of the ``BLK``.
        """
        styles = {}

        # the @fmx:TYPE is preserved in a @cals:rowstyle
        # the BLK level is also stored in this attribute
        blk_count = self._count_blk(fmx_ti_blk)
        blk_level = "TI.BLK-level{count}".format(count=blk_count)
        styles["rowstyle"] = blk_level

        styles.update(self.parse_cals_row_styles(fmx_ti_blk))

        return self._insert_blk_title_row(fmx_ti_blk, styles)

    def parse_fmx_sti_blk(self, fmx_sti_blk):
        """
        Parse a ``STI.BLK`` element, considering it like a row of a single cell.

        For instance:

        .. code-block:: xml

           <STI.BLK COL.START="1" COL.END="1">
             <P>STI.BLK title</P>
           </STI.BLK>

        :type  fmx_sti_blk: ElementType
        :param fmx_sti_blk: subtitle of the ``BLK``.
        """
        styles = {}

        # the @fmx:TYPE is preserved in a @cals:rowstyle
        # the BLK level is also stored in this attribute
        blk_count = self._count_blk(fmx_sti_blk)
        blk_level = "STI.BLK-level{count}".format(count=blk_count)
        styles["rowstyle"] = blk_level

        styles.update(self.parse_cals_row_styles(fmx_sti_blk))

        return self._insert_blk_title_row(fmx_sti_blk, styles)

    def parse_gr_notes(self, fmx_gr_notes):
        """
        Parse a ``GR.NOTES`` element, considering it like a row of a single cell.

        For instance:

        .. code-block:: xml

           <GR.NOTES>
             <TITLE>
               <TI>
                 <P>GR.NOTES Title</P>
               </TI>
             </TITLE>
             <NOTE NOTE.ID="N0001">
               <P>Table note</P>
             </NOTE>
           </GR.NOTES>

        :type  fmx_gr_notes: ElementType
        :param fmx_gr_notes: group of notes called in the table (``GR.NOTES``)

        .. versionchanged:: 0.5.1
           ``GR.NOTES`` elements can be embedded if the *embed_gr_notes* options is ``True``.
        """
        # -- Create a ROW
        state = self._state
        state.row = state.table.rows[state.row_pos]
        state.row.nature = "footer"

        styles = self.parse_cals_row_styles(fmx_gr_notes)

        # -- Create a CELL
        if self.embed_gr_notes:
            content = [fmx_gr_notes]
        elif self.contains_ie(fmx_gr_notes):
            content = ""
        else:
            text = [fmx_gr_notes.text] if fmx_gr_notes.text else []
            content = text + fmx_gr_notes.getchildren()

        state.row.insert_cell(
            content, styles=styles, nature=state.row.nature, width=state.table.bounding_box.width, height=1
        )

        return state

    def _count_blk(self, fmx_node):
        if self.formex_ns:
            blk_count = len(fmx_node.xpath("ancestor::fmx:BLK", namespaces={"fmx": self.formex_ns}))
        else:
            blk_count = len(fmx_node.xpath("ancestor::BLK"))
        return blk_count

    def _insert_blk_title_row(self, fmx_blk_title, styles):
        # -- Create a ROW
        state = self._state
        state.row = state.table.rows[state.row_pos]
        state.row.nature = "body"
        state.row.styles = styles

        fmx = self.get_formex_qname
        col_start = int(fmx_blk_title.attrib.get(fmx("COL.START"), "1"))
        col_end = int(fmx_blk_title.attrib.get(fmx("COL.END"), "1"))
        width = col_end - col_start + 1

        # -- Insert empty cells if necessary
        for x in range(1, col_start):
            state.row.insert_cell(None, nature=state.row.nature)

        # -- Create a CELL
        if self.contains_ie(fmx_blk_title):
            content = ""
        else:
            text = [fmx_blk_title.text] if fmx_blk_title.text else []
            content = text + fmx_blk_title.getchildren()
        state.row.insert_cell(content, width=width, height=1, nature=state.row.nature)

        return state

    def contains_ie(self, fmx_node):
        return bool(
            (self.formex_ns and fmx_node.xpath("fmx:IE", namespaces={"fmx": self.formex_ns}))
            or (not self.formex_ns and fmx_node.xpath("IE"))
        )

    def parse_fmx_cell(self, fmx_cell):
        """
        Parse a ``CELL`` element.

        :type  fmx_cell: ElementType
        :param fmx_cell: table cell

        .. versionchanged:: 0.5.2
           Add support for the ``@cals:cellstyle`` attribute (extension).
           This attribute is required for two-way conversion of Formex tables to CALS and vice versa.
           If the ``CELL/@TYPE`` and the ``ROW/@TYPE`` are different, we add a specific "cellstyle" style.
           This style will keep the ``CELL/@TYPE`` value.
        """
        fmx = self.get_formex_qname
        styles = {}

        # -- attribute @fmx:COL => not used

        # -- attribute @fmx:COLSPAN
        width = int(fmx_cell.attrib.get(fmx("COLSPAN"), "1"))

        # -- attribute @fmx:ROWSPAN
        height = int(fmx_cell.attrib.get(fmx("ROWSPAN"), "1"))

        # -- attribute @fmx:TYPE
        row_nature = self._state.row.nature
        type_map = {"ALIAS": "header", "HEADER": "header", "NORMAL": "body", "NOTCOL": "body", "TOTAL": "body"}
        cell_type = fmx_cell.attrib.get(fmx("TYPE"))
        cell_nature = type_map.get(cell_type)
        nature = cell_nature or row_nature

        # add the "cellstyle" if necessary
        fmx_row = fmx_cell.getparent()
        row_type = None if fmx_row is None else fmx_row.attrib.get(fmx("TYPE"))
        if (cell_type or "NORMAL") != (row_type or "NORMAL"):
            styles["cellstyle"] = cell_type

        # support for CALS-like elements and attributes
        cals = self.get_cals_qname

        # -- attribute @cals:colsep
        colsep = fmx_cell.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["x-cell-border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = fmx_cell.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["x-cell-border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:bgcolor
        bgcolor = fmx_cell.attrib.get(cals("bgcolor"))
        if bgcolor:
            styles["background-color"] = bgcolor

        # -- attributes @cals:namest and @cals:nameend
        name_start = fmx_cell.attrib.get(cals("namest"))
        name_end = fmx_cell.attrib.get(cals("nameend"))
        if name_start and name_end:
            col_start = int(re.findall(r"\d+", name_start)[0])
            col_end = int(re.findall(r"\d+", name_end)[0])
            width = col_end - col_start + 1

        # -- attribute @cals:morerows
        morerows = fmx_cell.attrib.get(cals("morerows"))
        if morerows:
            height = int(morerows) + 1

        # -- attribute @cals:valign
        valign = fmx_cell.attrib.get(cals("valign"))
        valign_map = {'top': 'top', 'middle': 'middle', 'bottom': 'bottom'}
        if valign in valign_map:
            # overrides parent's value
            styles["vertical-align"] = valign_map[valign]

        # -- attribute @cals:rowsep
        align = fmx_cell.attrib.get(cals("align"))
        align_map = {"left": "left", "right": "right", "center": "center", "justify": "justify", "char": "left"}
        if align in align_map:
            styles["align"] = align_map[align]

        # todo: calculate the ``@rotate`` attribute.

        # -- Create a CELL
        if self.contains_ie(fmx_cell):
            content = ""
        else:
            text = [fmx_cell.text] if fmx_cell.text else []
            content = text + fmx_cell.getchildren()
        self._state.row.insert_cell(content, width=width, height=height, styles=styles, nature=nature)

        return self._state  # mainly for unit test

    def parse_fmx_colspec(self, cals_colspec):
        """
        Parse a CALS-like ``colspec`` element.

        For instance:

        .. code-block:: xml

           <colspec
             colname="c1"
             colnum="1"
             colsep="1"
             rowsep="1"
             colwidth="30mm"
             align="center"/>

        :type  cals_colspec: ElementType
        :param cals_colspec: CALS-like ``colspec`` element.
        """
        cals = self.get_cals_qname
        styles = {}

        # -- attribute @cals:colname is ignored
        # -- attribute @cals:char is ignored
        # -- attribute @cals:charoff is ignored

        # -- attribute @cals:colnum
        colnum = cals_colspec.attrib.get(cals("colnum"))
        colnum = int(colnum) if colnum else self._state.col_pos

        # -- attribute @cals:colsep
        colsep = cals_colspec.attrib.get(cals("colsep"))
        colsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if colsep in colsep_map:
            styles["border-right"] = colsep_map[colsep]

        # -- attribute @cals:rowsep
        rowsep = cals_colspec.attrib.get(cals("rowsep"))
        rowsep_map = {"0": BORDER_NONE, "1": BORDER_SOLID}
        if rowsep in rowsep_map:
            styles["border-bottom"] = rowsep_map[rowsep]

        # -- attribute @cals:rowsep
        colwidth = cals_colspec.attrib.get(cals("colwidth"))
        if colwidth:
            styles["width"] = colwidth

        # -- attribute @cals:rowsep
        align = cals_colspec.attrib.get(cals("align"))
        align_map = {"left": "left", "right": "right", "center": "center", "justify": "justify", "char": "left"}
        if align in align_map:
            styles["align"] = align_map[align]

        state = self._state
        state.col = state.table.cols[colnum]
        state.col.styles.update(styles)

        return state



# coding: utf-8
"""
Formex 4 to CALS converter
==========================

.. versionadded:: 0.5.0
"""


from benker.builders.cals import CalsBuilder
from benker.converters.base_converter import BaseConverter
from benker.parsers.formex import FormexParser


class Formex2CalsConverter(BaseConverter):
    """
    Formex 4 to CALS converter
    """

    parser_cls = FormexParser
    builder_cls = CalsBuilder


def convert_formex2cals(src_xml, dst_xml, **options):
    """
    Convert Formex 4 tables to Cals tables.

    :param str src_xml:
        Source path of the XML file to convert.

    :param str dst_xml:
        Destination path of the XML file to produce.

    :keyword options:
        Dictionary of parsing/building options.

        **Common parsing options:**

        ``encoding`` (default: "utf-8"):
            XML encoding of the destination file.

        **Formex parser options:**

        ``formex_ns`` (default ``None``):
            Namespace to use for Formex elements and attributes parsing.
            Set ``None`` (or "") if you don't use namespace.

        ``cals_ns`` (default ``None``):
            Namespace to use for CALS-like elements and attributes parsing.
            Set ``None`` (or "") if you don't use namespace.

        ``embed_gr_notes`` (default ``False``):
            If ``True``, Embed the ``GR.NOTES`` in a row/cell, else only copy the content (not the ``GR.NOTES`` tag).

        **CALS builder options:**

        ``cals_ns`` (default: ``None``):
            Namespace to use for CALS-like elements and attributes to generate.
            Set ``None`` (or "") if you don't want to use namespace.

        ``cals_prefix`` (default: ``None``):
            Namespace prefix to use for CALS-like elements and attributes to generate.

        ``width_unit`` (default: "mm"):
            Unit to use for column widths.
            Possible values are: 'cm', 'dm', 'ft', 'in', 'm', 'mm', 'pc', 'pt', 'px'.

        ``table_in_tgroup`` (default: ``False``):
            Where should we put the table properties:

            -   ``False`` to insert the attributes ``@colsep``, ``@rowsep``,
                and ``@tabstyle`` in the ``<table>`` element,

            -   ``True`` to insert the attributes ``@colsep``, ``@rowsep``,
                and ``@tgroupstyle`` in the ``<tgroup>`` element.

        ``tgroup_sorting`` (default: ``["header", "footer", "body"]``):
            List used to sort (and group) the rows in a ``tgroup``.
            The sorting is done according to the row natures
            which is by default: ``["header", "footer", "body"]``
            (this order match the CALS DTD defaults,
            where the footer is between the header and the body.
            To move the footer to the end, you can use ``["header", "body", "footer"]``.

    .. versionchanged:: 0.5.0
       Add the options *cals_ns*, *cals_prefix*, *tgroup_sorting*.
    """
    converter = Formex2CalsConverter()
    converter.convert_file(src_xml, dst_xml, **options)



# coding: utf-8
"""
Office Open XML parser
======================

This module can parse Office Open XML (OOXML) tables.

Specifications:

- The documentation about OOXML Table is available online at
  `Word Processing - Table Grid/Column Definition <http://officeopenxml.com/WPtableGrid.php>`_.
"""
import collections

from lxml import etree

from benker.box import Box
from benker.cell import get_content_text
from benker.common.lxml_iterwalk import iterwalk
from benker.parsers.base_parser import BaseParser
from benker.parsers.ooxml.namespaces import NS
from benker.parsers.ooxml.namespaces import value_of
from benker.parsers.ooxml.namespaces import w
from benker.parsers.ooxml.w_pg_sz import PgSz
from benker.parsers.ooxml.w_shd import Shd
from benker.table import Table

#: See w:ST_Border: http://www.datypic.com/sc/ooxml/t-w_ST_Border.html
#: See CSS styles: https://www.w3.org/wiki/CSS/Properties/border-top-style
# fmt: off
_BORDER_STYLE_MAPPING = {
    # fmt: off
    # No Border
    'nil': 'none',

    # No Border
    'none': 'none',

    # Single Line Border
    'single': 'solid',

    # Single Line Border
    'thick': 'w-thick',

    # Double Line Border
    'double': 'double',

    # Dotted Line Border
    'dotted': 'dotted',

    # Dashed Line Border
    'dashed': 'dashed',

    # Dot Dash Line Border
    'dotDash': 'w-dot-dash',

    # Dot Dot Dash Line Border
    'dotDotDash': 'w-dot-dot-dash',

    # Triple Line Border
    'triple': 'w-triple',

    # Thin, Thick Line Border
    'thinThickSmallGap': 'w-thin-thick-small-gap',

    # Thick, Thin Line Border
    'thickThinSmallGap': 'w-thick-thin-small-gap',

    # Thin, Thick, Thin Line Border
    'thinThickThinSmallGap': 'w-thin-thick-thin-small-gap',

    # Thin, Thick Line Border
    'thinThickMediumGap': 'w-thin-thick-medium-gap',

    # Thick, Thin Line Border
    'thickThinMediumGap': 'w-thick-thin-medium-gap',

    # Thin, Thick, Thin Line Border
    'thinThickThinMediumGap': 'w-thin-thick-thin-medium-gap',

    # Thin, Thick Line Border
    'thinThickLargeGap': 'w-thin-thick-large-gap',

    # Thick, Thin Line Border
    'thickThinLargeGap': 'w-thick-thin-large-gap',

    # Thin, Thick, Thin Line Border
    'thinThickThinLargeGap': 'w-thin-thick-thin-large-gap',

    # Wavy Line Border
    'wave': 'w-wave',

    # Double Wave Line Border
    'doubleWave': 'w-double-wave',

    # Dashed Line Border
    'dashSmallGap': 'w-dash-small-gap',

    # Dash Dot Strokes Line Border
    'dashDotStroked': 'w-dash-dot-stroked',

    # 3D Embossed Line Border
    'threeDEmboss': 'w-three-d-emboss',

    # 3D Engraved Line Border
    'threeDEngrave': 'w-three-d-engrave',

    # Outset Line Border
    'outset': 'outset',

    # Inset Line Border
    'inset': 'inset',

    # Apples Art Border
    'apples': 'w-apples',

    # Arched Scallops Art Border
    'archedScallops': 'w-arched-scallops',

    # Baby Pacifier Art Border
    'babyPacifier': 'w-baby-pacifier',

    # Baby Rattle Art Border
    'babyRattle': 'w-baby-rattle',

    # Three Color Balloons Art Border
    'balloons3Colors': 'w-balloons3-colors',

    # Hot Air Balloons Art Border
    'balloonsHotAir': 'w-balloons-hot-air',

    # Black Dash Art Border
    'basicBlackDashes': 'w-basic-black-dashes',

    # Black Dot Art Border
    'basicBlackDots': 'w-basic-black-dots',

    # Black Square Art Border
    'basicBlackSquares': 'w-basic-black-squares',

    # Thin Line Art Border
    'basicThinLines': 'w-basic-thin-lines',

    # White Dash Art Border
    'basicWhiteDashes': 'w-basic-white-dashes',

    # White Dot Art Border
    'basicWhiteDots': 'w-basic-white-dots',

    # White Square Art Border
    'basicWhiteSquares': 'w-basic-white-squares',

    # Wide Inline Art Border
    'basicWideInline': 'w-basic-wide-inline',

    # Wide Midline Art Border
    'basicWideMidline': 'w-basic-wide-midline',

    # Wide Outline Art Border
    'basicWideOutline': 'w-basic-wide-outline',

    # Bats Art Border
    'bats': 'w-bats',

    # Birds Art Border
    'birds': 'w-birds',

    # Birds Flying Art Border
    'birdsFlight': 'w-birds-flight',

    # Cabin Art Border
    'cabins': 'w-cabins',

    # Cake Art Border
    'cakeSlice': 'w-cake-slice',

    # Candy Corn Art Border
    'candyCorn': 'w-candy-corn',

    # Knot Work Art Border
    'celticKnotwork': 'w-celtic-knotwork',

    # Certificate Banner Art Border
    'certificateBanner': 'w-certificate-banner',

    # Chain Link Art Border
    'chainLink': 'w-chain-link',

    # Champagne Bottle Art Border
    'champagneBottle': 'w-champagne-bottle',

    # Black and White Bar Art Border
    'checkedBarBlack': 'w-checked-bar-black',

    # Color Checked Bar Art Border
    'checkedBarColor': 'w-checked-bar-color',

    # Checkerboard Art Border
    'checkered': 'w-checkered',

    # Christmas Tree Art Border
    'christmasTree': 'w-christmas-tree',

    # Circles And Lines Art Border
    'circlesLines': 'w-circles-lines',

    # Circles and Rectangles Art Border
    'circlesRectangles': 'w-circles-rectangles',

    # Wave Art Border
    'classicalWave': 'w-classical-wave',

    # Clocks Art Border
    'clocks': 'w-clocks',

    # Compass Art Border
    'compass': 'w-compass',

    # Confetti Art Border
    'confetti': 'w-confetti',

    # Confetti Art Border
    'confettiGrays': 'w-confetti-grays',

    # Confetti Art Border
    'confettiOutline': 'w-confetti-outline',

    # Confetti Streamers Art Border
    'confettiStreamers': 'w-confetti-streamers',

    # Confetti Art Border
    'confettiWhite': 'w-confetti-white',

    # Corner Triangle Art Border
    'cornerTriangles': 'w-corner-triangles',

    # Dashed Line Art Border
    'couponCutoutDashes': 'w-coupon-cutout-dashes',

    # Dotted Line Art Border
    'couponCutoutDots': 'w-coupon-cutout-dots',

    # Maze Art Border
    'crazyMaze': 'w-crazy-maze',

    # Butterfly Art Border
    'creaturesButterfly': 'w-creatures-butterfly',

    # Fish Art Border
    'creaturesFish': 'w-creatures-fish',

    # Insects Art Border
    'creaturesInsects': 'w-creatures-insects',

    # Ladybug Art Border
    'creaturesLadyBug': 'w-creatures-lady-bug',

    # Cross-stitch Art Border
    'crossStitch': 'w-cross-stitch',

    # Cupid Art Border
    'cup': 'w-cup',

    # Archway Art Border
    'decoArch': 'w-deco-arch',

    # Color Archway Art Border
    'decoArchColor': 'w-deco-arch-color',

    # Blocks Art Border
    'decoBlocks': 'w-deco-blocks',

    # Gray Diamond Art Border
    'diamondsGray': 'w-diamonds-gray',

    # Double D Art Border
    'doubleD': 'w-double-d',

    # Diamond Art Border
    'doubleDiamonds': 'w-double-diamonds',

    # Earth Art Border
    'earth1': 'w-earth1',

    # Earth Art Border
    'earth2': 'w-earth2',

    # Shadowed Square Art Border
    'eclipsingSquares1': 'w-eclipsing-squares1',

    # Shadowed Square Art Border
    'eclipsingSquares2': 'w-eclipsing-squares2',

    # Painted Egg Art Border
    'eggsBlack': 'w-eggs-black',

    # Fans Art Border
    'fans': 'w-fans',

    # Film Reel Art Border
    'film': 'w-film',

    # Firecracker Art Border
    'firecrackers': 'w-firecrackers',

    # Flowers Art Border
    'flowersBlockPrint': 'w-flowers-block-print',

    # Daisy Art Border
    'flowersDaisies': 'w-flowers-daisies',

    # Flowers Art Border
    'flowersModern1': 'w-flowers-modern1',

    # Flowers Art Border
    'flowersModern2': 'w-flowers-modern2',

    # Pansy Art Border
    'flowersPansy': 'w-flowers-pansy',

    # Red Rose Art Border
    'flowersRedRose': 'w-flowers-red-rose',

    # Roses Art Border
    'flowersRoses': 'w-flowers-roses',

    # Flowers in a Teacup Art Border
    'flowersTeacup': 'w-flowers-teacup',

    # Small Flower Art Border
    'flowersTiny': 'w-flowers-tiny',

    # Gems Art Border
    'gems': 'w-gems',

    # Gingerbread Man Art Border
    'gingerbreadMan': 'w-gingerbread-man',

    # Triangle Gradient Art Border
    'gradient': 'w-gradient',

    # Handmade Art Border
    'handmade1': 'w-handmade1',

    # Handmade Art Border
    'handmade2': 'w-handmade2',

    # Heart-Shaped Balloon Art Border
    'heartBalloon': 'w-heart-balloon',

    # Gray Heart Art Border
    'heartGray': 'w-heart-gray',

    # Hearts Art Border
    'hearts': 'w-hearts',

    # Pattern Art Border
    'heebieJeebies': 'w-heebie-jeebies',

    # Holly Art Border
    'holly': 'w-holly',

    # House Art Border
    'houseFunky': 'w-house-funky',

    # Circular Art Border
    'hypnotic': 'w-hypnotic',

    # Ice Cream Cone Art Border
    'iceCreamCones': 'w-ice-cream-cones',

    # Light Bulb Art Border
    'lightBulb': 'w-light-bulb',

    # Lightning Art Border
    'lightning1': 'w-lightning1',

    # Lightning Art Border
    'lightning2': 'w-lightning2',

    # Map Pins Art Border
    'mapPins': 'w-map-pins',

    # Maple Leaf Art Border
    'mapleLeaf': 'w-maple-leaf',

    # Muffin Art Border
    'mapleMuffins': 'w-maple-muffins',

    # Marquee Art Border
    'marquee': 'w-marquee',

    # Marquee Art Border
    'marqueeToothed': 'w-marquee-toothed',

    # Moon Art Border
    'moons': 'w-moons',

    # Mosaic Art Border
    'mosaic': 'w-mosaic',

    # Musical Note Art Border
    'musicNotes': 'w-music-notes',

    # Patterned Art Border
    'northwest': 'w-northwest',

    # Oval Art Border
    'ovals': 'w-ovals',

    # Package Art Border
    'packages': 'w-packages',

    # Black Palm Tree Art Border
    'palmsBlack': 'w-palms-black',

    # Color Palm Tree Art Border
    'palmsColor': 'w-palms-color',

    # Paper Clip Art Border
    'paperClips': 'w-paper-clips',

    # Papyrus Art Border
    'papyrus': 'w-papyrus',

    # Party Favor Art Border
    'partyFavor': 'w-party-favor',

    # Party Glass Art Border
    'partyGlass': 'w-party-glass',

    # Pencils Art Border
    'pencils': 'w-pencils',

    # Character Art Border
    'people': 'w-people',

    # Waving Character Border
    'peopleWaving': 'w-people-waving',

    # Character With Hat Art Border
    'peopleHats': 'w-people-hats',

    # Poinsettia Art Border
    'poinsettias': 'w-poinsettias',

    # Postage Stamp Art Border
    'postageStamp': 'w-postage-stamp',

    # Pumpkin Art Border
    'pumpkin1': 'w-pumpkin1',

    # Push Pin Art Border
    'pushPinNote2': 'w-push-pin-note2',

    # Push Pin Art Border
    'pushPinNote1': 'w-push-pin-note1',

    # Pyramid Art Border
    'pyramids': 'w-pyramids',

    # Pyramid Art Border
    'pyramidsAbove': 'w-pyramids-above',

    # Quadrants Art Border
    'quadrants': 'w-quadrants',

    # Rings Art Border
    'rings': 'w-rings',

    # Safari Art Border
    'safari': 'w-safari',

    # Saw tooth Art Border
    'sawtooth': 'w-sawtooth',

    # Gray Saw tooth Art Border
    'sawtoothGray': 'w-sawtooth-gray',

    # Scared Cat Art Border
    'scaredCat': 'w-scared-cat',

    # Umbrella Art Border
    'seattle': 'w-seattle',

    # Shadowed Squares Art Border
    'shadowedSquares': 'w-shadowed-squares',

    # Shark Tooth Art Border
    'sharksTeeth': 'w-sharks-teeth',

    # Bird Tracks Art Border
    'shorebirdTracks': 'w-shorebird-tracks',

    # Rocket Art Border
    'skyrocket': 'w-skyrocket',

    # Snowflake Art Border
    'snowflakeFancy': 'w-snowflake-fancy',

    # Snowflake Art Border
    'snowflakes': 'w-snowflakes',

    # Sombrero Art Border
    'sombrero': 'w-sombrero',

    # Southwest-themed Art Border
    'southwest': 'w-southwest',

    # Stars Art Border
    'stars': 'w-stars',

    # Stars On Top Art Border
    'starsTop': 'w-stars-top',

    # 3-D Stars Art Border
    'stars3d': 'w-stars3d',

    # Stars Art Border
    'starsBlack': 'w-stars-black',

    # Stars With Shadows Art Border
    'starsShadowed': 'w-stars-shadowed',

    # Sun Art Border
    'sun': 'w-sun',

    # Whirligig Art Border
    'swirligig': 'w-swirligig',

    # Torn Paper Art Border
    'tornPaper': 'w-torn-paper',

    # Black Torn Paper Art Border
    'tornPaperBlack': 'w-torn-paper-black',

    # Tree Art Border
    'trees': 'w-trees',

    # Triangle Art Border
    'triangleParty': 'w-triangle-party',

    # Triangles Art Border
    'triangles': 'w-triangles',

    # Tribal Art Border One
    'tribal1': 'w-tribal1',

    # Tribal Art Border Two
    'tribal2': 'w-tribal2',

    # Tribal Art Border Three
    'tribal3': 'w-tribal3',

    # Tribal Art Border Four
    'tribal4': 'w-tribal4',

    # Tribal Art Border Five
    'tribal5': 'w-tribal5',

    # Tribal Art Border Six
    'tribal6': 'w-tribal6',

    # Twisted Lines Art Border
    'twistedLines1': 'w-twisted-lines1',

    # Twisted Lines Art Border
    'twistedLines2': 'w-twisted-lines2',

    # Vine Art Border
    'vine': 'w-vine',

    # Wavy Line Art Border
    'waveline': 'w-waveline',

    # Weaving Angles Art Border
    'weavingAngles': 'w-weaving-angles',

    # Weaving Braid Art Border
    'weavingBraid': 'w-weaving-braid',

    # Weaving Ribbon Art Border
    'weavingRibbon': 'w-weaving-ribbon',

    # Weaving Strips Art Border
    'weavingStrips': 'w-weaving-strips',

    # White Flowers Art Border
    'whiteFlowers': 'w-white-flowers',

    # Woodwork Art Border
    'woodwork': 'w-woodwork',

    # Crisscross Art Border
    'xIllusions': 'w-x-illusions',

    # Triangle Art Border
    'zanyTriangles': 'w-zany-triangles',

    # Zigzag Art Border
    'zigZag': 'w-zig-zag',

    # Zigzag stitch
    'zigZagStitch': 'w-zig-zag-stitch',
    # fmt: on
}
# fmt: on


def _get_border_properties(w_tbl_borders, style_xpath_mapping):
    # - Get the cell properties for each direction: 'top', 'right'...
    #   Values are converted to HTML values, size are in 'pt'
    properties = []
    for style, xpath in style_xpath_mapping:
        prop = {}
        color = value_of(w_tbl_borders, xpath.format(attr='color'))
        if color and color != "auto":
            prop['color'] = "#" + color
        shadow = value_of(w_tbl_borders, xpath.format(attr='shadow'))
        if shadow:
            prop['shadow'] = {"true": True, "false": False}[shadow]
        space = value_of(w_tbl_borders, xpath.format(attr='space'))
        if space:
            # unit is 'pt'
            prop['space'] = float(space)
        sz = value_of(w_tbl_borders, xpath.format(attr='sz'))
        if sz:
            # convert eighths of a point to 'pt'
            prop['sz'] = float(sz) / 8
        val = value_of(w_tbl_borders, xpath.format(attr='val'))
        if val:
            val = "none" if val == "nil" else val  # "nil" is "none" -- no border
            prop['val'] = _BORDER_STYLE_MAPPING.get(val, 'w-' + val)
        properties.append((style, prop))
    return properties


def _border_properties_to_styles(properties):
    """
    Convert the properties to border styles.

    :param properties: Ordered list of OOXML properties

    :return: dictionary of border styles
    """
    styles = {}

    # - 'border-top', 'border-right', 'border-bottom', 'border-left'
    for style, prop in properties:
        # formats -- order is important
        formats = ('val', "{val}"), ('sz', "{sz}pt"), ('color', "{color}")
        values = [fmt.format(**prop) for key, fmt in formats if key in prop]
        if values:
            styles[style] = " ".join(values)

    if any('space' in prop for style, prop in properties):
        # - 'border-collapse' property selects a table's border model(separated or collapsing).
        #   Value is "collapse" if all spaces == 0
        spaces = [prop.get('space', 0) for style, prop in properties]
        all_spaces_are_nul = all(space == 0 for space in spaces)
        styles['border-collapse'] = "collapse" if all_spaces_are_nul else "separate"

        # - 'border-spacing' property specifies the distance between the borders of adjacent cells.
        if not all_spaces_are_nul:
            spacing = ["{0}pt".format(prop.get('space', 0)) for style, prop in properties]
            styles['border-spacing'] = " ".join(spacing)

    # - The box-shadow property attaches one or more shadows to an element.
    #   Use the border size, the default color, without effect (blur...)
    has_shadow = any(prop.get('shadow') for style, prop in properties)
    if has_shadow:
        shadow = ["{0}pt".format(prop['sz'] if prop.get('shadow') else "0pt") for style, prop in properties]
        styles['box-shadow'] = " ".join(shadow)

    return styles


def _get_table_borders(w_tbl_borders):
    """
    Construct a border dictionary from the ``w:tblPr/w:tblBorders`` properties.

    See `Table Properties <http://officeopenxml.com/WPtableProperties.php>`_
    and `Table Borders <http://officeopenxml.com/WPtableBorders.php>`_.

    :param w_tbl_borders:
        Properties of the OOXML table.

    :rtype: benker.parsers.ooxml.OoxmlBorder
    :return: New instance.
    """
    # Table Cell Top Right to Bottom Left Diagonal Border
    #  http://www.datypic.com/sc/ooxml/e-w_tr2bl-1.html
    #  xpath='w:tcPr/w:tcBorders/w:tr2bl'
    #  ex.: <w:tr2bl w:val="single" w:sz="4" w:space="0" w:color="auto"/>

    # Table Cell Top Left to Bottom Right Diagonal Border
    #  http://www.datypic.com/sc/ooxml/e-w_tl2br-1.html
    #  xpath='w:tcPr/w:tcBorders/w:tl2br'
    #  ex.: <w:tl2br w:val="single" w:sz="4" w:space="0" w:color="auto"/>

    if w_tbl_borders is None:
        return {}
    # style_xpath_mapping -- order is important
    style_xpath_mapping = [
        ('border-top', "w:top/@w:{attr}"),
        ('border-right', "w:end/@w:{attr} | w:right/@w:{attr}"),
        ('border-bottom', "w:bottom/@w:{attr}"),
        ('border-left', "w:start/@w:{attr} | w:left/@w:{attr}"),
        ('x-border-tr2bl', "w:tr2bl/@w:{attr}"),
        ('x-border-tl2br', "w:tl2br/@w:{attr}"),
    ]
    properties = _get_border_properties(w_tbl_borders, style_xpath_mapping)
    styles = _border_properties_to_styles(properties)
    return styles


def _get_cell_borders(w_tbl_borders):
    """
    Construct a border dictionary from the ``w:tblPr/w:tblBorders`` properties.

    See `Table Properties <http://officeopenxml.com/WPtableProperties.php>`_
    and `Table Borders <http://officeopenxml.com/WPtableBorders.php>`_.

    :param w_tbl_borders:
        Properties of the OOXML table.

    :rtype: benker.parsers.ooxml.OoxmlBorder
    :return: New instance.
    """
    if w_tbl_borders is None:
        return {}
    # style_xpath_mapping -- order is important
    style_xpath_mapping = [
        ('border-top', "w:insideH/@w:{attr}"),
        ('border-right', "w:insideV/@w:{attr}"),
        ('border-bottom', "w:insideH/@w:{attr}"),
        ('border-left', "w:insideV/@w:{attr}"),
    ]
    properties = _get_border_properties(w_tbl_borders, style_xpath_mapping)
    styles = _border_properties_to_styles(properties)
    return styles


def _get_style_borders(w_styles, style_id):
    if w_styles is None or style_id is None:
        return {}
    w_style = value_of(w_styles, 'w:style[@w:styleId = "{0}"]'.format(style_id))
    if w_style is None:
        return {}

    # - get parent styles (if it exist)
    based_on_id = value_of(w_style, 'w:basedOn/@w:val')
    parent_styles = _get_style_borders(w_styles, based_on_id)

    # - get child styles
    w_tbl_borders = value_of(w_style, 'w:tblPr/w:tblBorders')
    table_borders = _get_table_borders(w_tbl_borders)
    cell_borders = _get_cell_borders(w_tbl_borders)
    child_styles = table_borders.copy()
    child_styles.update({'x-cell-' + key: value for key, value in cell_borders.items()})

    # - *child_styles* override *parent_styles*
    real_styles = parent_styles.copy()
    real_styles.update({key: value for key, value in child_styles.items() if value is not None})
    return real_styles


class OoxmlParser(BaseParser):
    """
    Office Open XML parser.
    """

    def __init__(self, builder, styles_path=None, **options):
        """
        Construct a parser

        :type  builder: benker.builders.base_builder.BaseBuilder
        :param builder:
            Builder used by this parser to instantiate :class:`~benker.table.Table` objects.

        :param str styles_path:
            Path to the stylesheet to use to resole table styles.
            In an uncompressed ``.docx`` tree structure, the stylesheet path
            is ``word/styles.xml``.

        :param str options: Extra conversion options.
            See :meth:`~benker.converters.base_converter.BaseConverter.convert_file`
            to have a list of all possible options.
        """
        self._w_styles = None
        self.styles_path = styles_path
        super(OoxmlParser, self).__init__(builder, **options)

    def transform_tables(self, tree):
        self._w_styles = etree.parse(self.styles_path) if self.styles_path else None
        self._w_styles = self._w_styles or value_of(tree, ".//w:styles")

        for w_tbl in tree.xpath("//w:tbl", namespaces=NS):
            table = self.parse_table(w_tbl)
            table_elem = self.builder.generate_table_tree(table)
            parent = w_tbl.getparent()
            index = parent.index(w_tbl)
            parent.insert(index, table_elem)
            table_elem.tail = w_tbl.tail
            parent.remove(w_tbl)

    def parse_table(self, w_tbl):
        """
        Convert a Office Open XML ``<w:tbl>`` into CALS ``<table>``

        :type  w_tbl: etree._Element
        :param w_tbl: Office Open XML element.

        :rtype: etree._Element
        :return: CALS element.
        """
        state = self._state
        state.reset()

        elements = {w(name) for name in {'tbl', 'tblGrid', 'gridCol', 'tr', 'tc'}}
        context = iterwalk(w_tbl, events=('start', 'end'), tag=elements)

        depth = 0
        for action, elem in context:
            elem_tag = elem.tag
            if elem_tag == w('tbl'):
                if action == 'start':
                    depth += 1
                else:
                    depth -= 1
            if depth > 1:
                # .. note:: context.skip_subtree() is not available for all version of lxml
                # This <tbl> element is inside the table.
                # It will be handled separately in another call to convert_tbl()
                continue
            if action == 'start':
                if elem_tag == w('tbl'):
                    self.parse_tbl(elem)

                elif elem_tag == w('tblGrid'):
                    # this element has no specific data
                    pass

                elif elem_tag == w('gridCol'):
                    state.next_col()
                    self.parse_grid_col(elem)

                elif elem_tag == w('tr'):
                    state.next_row()
                    self.parse_tr(elem)

                elif elem_tag == w('tc'):
                    state.next_col()
                    self.parse_tc(elem)

                else:
                    raise NotImplementedError(elem_tag)
            else:
                if elem_tag == w('tr'):
                    # add missing entries
                    bounding_box = Box(1, state.row_pos, len(state.table.cols), state.row_pos)
                    state.table.fill_missing(bounding_box, None, nature=state.row.nature)

        return state.table

    def parse_tbl(self, w_tbl):
        """
        Parse a ``<w:tbl>`` element.

        See: `Table Properties <http://officeopenxml.com/WPtableProperties.php>`_.

        :type  w_tbl: etree._Element
        :param w_tbl: Table element.

        .. versionchanged:: 0.4.0
           The section width and height are now stored in the 'x-sect-size' table style (units in 'pt').
        """
        style_id = value_of(w_tbl, "w:tblPr/w:tblStyle/@w:val")

        # - Table and borders are extracted from the style (if possible)
        #   and then from the ``w:tblPr/w:tblBorders`` properties.

        style_borders = _get_style_borders(self._w_styles, style_id)
        w_tbl_borders = value_of(w_tbl, 'w:tblPr/w:tblBorders')

        # - Table borders (frame) and Cell borders (colsep/rowsep) use the "x-cell-" prefix

        table_borders = _get_table_borders(w_tbl_borders)
        real_table_borders = style_borders.copy()
        real_table_borders.update({key: value for key, value in table_borders.items() if value is not None})

        attrs = real_table_borders.copy()

        # -- Table shading
        shd = Shd(value_of(w_tbl, 'w:tblPr/w:shd'))
        attrs.update(shd.styles)

        # -- Sections: http://officeopenxml.com/WPsection.php

        # A section's properties are stored in a sectPr element.
        # For all sections except the last section, the sectPr element is stored as
        # a child element of the last paragraph in the section. For the last section,
        # the sectPr is stored as a child element of the body element.

        w_sect_pr = value_of(w_tbl, 'following::w:p/w:pPr/w:sectPr | following::w:sectPr')

        pg_sz = PgSz(value_of(w_sect_pr, 'w:pgSz'))
        attrs.update(pg_sz.styles)

        # - w:cols -- Specifies the set of columns for the section.
        # - ``x-sect-cols``: Section column number
        #   Default value is "1" -- useful for @pgwide
        sect_cols = value_of(w_sect_pr, 'w:cols/@w:num')
        if sect_cols is None:
            if w_sect_pr is None:
                sect_cols = "1"  # type: str
            else:
                sect_cols = w_sect_pr.xpath('count(w:cols/w:col)', namespaces=NS)  # type: float
                sect_cols = str(int(sect_cols)) if sect_cols else "1"  # type: str
        attrs['x-sect-cols'] = sect_cols

        # - The HTML ``class`` attribute is not a regular style.
        #   We use the table ``nature``instead.

        self._state.table = Table(styles=attrs, nature=style_id)

    def parse_grid_col(self, w_grid_col):
        """
        Parse a ``<w:gridCol>`` element.

        See: `Table Grid/Column Definition <http://officeopenxml.com/WPtableGrid.php>`_.

        :type  w_grid_col: etree._Element
        :param w_grid_col: Table element.
        """
        # w:w => width of the column in twentieths of a point.
        width = float(w_grid_col.attrib[w('w')]) / 20  # pt
        state = self._state
        styles = {u"width": u"{0:0.2f}pt".format(width)}
        state.col = state.table.cols[state.col_pos]
        state.col.styles.update(styles)

    def parse_tr(self, w_tr):
        """
        Parse a ``<w:tr>`` element.

        See: `Table Row Properties <http://officeopenxml.com/WPtableRowProperties.php>`_.

        :type  w_tr: etree._Element
        :param w_tr: Table element.
        """

        # - w:tblHeader => the current row should be repeated at the top
        #   of each new page on which the table is displayed.
        #   This is a simple boolean property, so you can specify a val attribute of true or false.
        #
        #   <w:trPr>
        #     <w:tblHeader/>
        #   </w:trPr>
        #
        w_tbl_header = value_of(w_tr, "w:trPr/w:tblHeader")
        if w_tbl_header is not None:
            w_tbl_header = value_of(w_tr, "w:trPr/w:tblHeader/@w:val", default=u"true")
        nature = {"true": u"header", "false": u"body", None: u"body"}[w_tbl_header]
        state = self._state
        state.row = state.table.rows[state.row_pos]
        state.row.nature = nature

        # - w:trHeight => height of the row
        #
        #   <w:trPr>
        #     <w:trHeight w:val="567"/>
        #   </w:trPr>
        #
        w_tr_height = value_of(w_tr, "w:trPr/w:trHeight")
        if w_tr_height is not None:
            h_rule = value_of(w_tr, "w:trPr/w:tblHeader/@w:hRule", default="auto")
            # Possible values are:
            # - atLeast (height should be at least the value specified),
            # - exact (height should be exactly the value specified), or
            # - auto (height is determined based on the height of the contents, so the value is ignored).
            style = {'atLeast': u'min-height', 'exact': u'height', 'auto': None}[h_rule]
            if style:
                val = value_of(w_tr, "w:trPr/w:tblHeader/@w:val", default="0")
                # Specifies the row's height, in twentieths of a point.
                height = float(val) / 20  # pt
                state.row.styles[style] = "{0:0.2f}pt".format(height)

        # - w:ins => revision marks: A row can be marked as "inserted".
        #
        #   <w:trPr>
        #     <w:ins w:id="0" w:author="Laurent Laporte" w:date="2018-11-21T18:08:00Z"/>
        #   </w:trPr>
        #
        w_ins = value_of(w_tr, "w:trPr/w:ins")
        if w_ins is not None:
            state.row.styles['x-ins'] = True
            style_xpath_mapping = [
                ('x-ins-id', "w:trPr/w:ins/@w:id"),
                ('x-ins-author', "w:trPr/w:ins/@w:author"),
                ('x-ins-date', "w:trPr/w:ins/@w:date"),
            ]
            for style, xpath in style_xpath_mapping:
                value = value_of(w_tr, xpath)
                if value:
                    state.row.styles[style] = value

        # note: ``vertical-align`` attribute is not available for a row => see w:tcPr instead

    def parse_tc(self, w_tc):
        """
        Parse a ``<w:tc>`` element.

        See: `Table Cell Properties <http://officeopenxml.com/WPtableCellProperties.php>`_.

        :type  w_tc: etree._Element
        :param w_tc: Table element.

        .. versionchanged:: 0.4.4
           Improved empty cells detection for Formex4 conversion (``<IE/>`` tag management).

        .. versionchanged:: 0.5.1
           XML indentation between cell paragraphs is ignored.
        """
        state = self._state

        # w:gridSpan => number of logical columns across which the cell spans
        width = int(value_of(w_tc, "w:tcPr/w:gridSpan/@w:val", default=u"1"))

        # take the colspan into account:
        state.col_pos += width - 1

        # w:vMerge => specifies that the cell is part of a vertically merged set of cells.
        w_v_merge = value_of(w_tc, "w:tcPr/w:vMerge")
        if w_v_merge is not None:
            w_v_merge = value_of(w_tc, "w:tcPr/w:vMerge/@w:val", default=u"continue")
        if w_v_merge is None:
            # no merge
            height = 1
        elif w_v_merge == u"continue":
            # the current cell continues a previously existing merge group
            state.table.expand((state.col_pos, state.row.row_pos - 1), height=1)
            height = None
        elif w_v_merge == u"restart":
            # the current cell starts a new merge group
            height = 1
        else:
            raise NotImplementedError(w_v_merge)

        if height:
            styles = {}

            # -- Cell shading
            shd = Shd(value_of(w_tc, 'w:tcPr/w:shd'))
            styles.update(shd.styles)

            # -- Vertical alignment
            #
            # w:vAlign => Specifies the vertical alignment for text between the top and bottom margins of the cell.
            #
            # Possible values are:
            # - bottom - Specifies that the text should be vertically aligned to the bottom margin.
            # - center - Specifies that the text should be vertically aligned to the center of the cell.
            # - top - Specifies that the text should be vertically aligned to the top margin.
            w_v_align = value_of(w_tc, "w:tcPr/w:vAlign")
            if w_v_align is not None:
                w_v_align = value_of(w_tc, "w:tcPr/w:vAlign/@w:val", default=u"top")
                # CSS/Properties/vertical-align
                # valid values: http://www.datypic.com/sc/ooxml/t-w_ST_VerticalJc.html
                # fmt: off
                v_align = {
                    "top": u"top",
                    "center": u"middle",
                    "bottom": u"bottom",
                    "both": u"w-both",
                }[w_v_align]
                # fmt: on
                styles["vertical-align"] = v_align

            # -- Horizontal alignment
            #
            # Horizontal alignment is done at paragraph level, inside the cell.
            # We can calculate the cell alignment base on the paragraph properties,
            # for instance ``<w:p><w:pPr><w:jc w:val="right"/>``,
            # see: http://officeopenxml.com/WPalignment.php
            #
            # We use the most common alignment for cell alignment.
            w_p_list = w_tc.xpath("w:p", namespaces=NS)
            w_jc_counter = collections.Counter(value_of(w_p, "w:pPr/w:jc/@w:val") for w_p in w_p_list)
            w_jc = w_jc_counter.most_common(1)[0][0]  # type: str or None
            if w_jc is not None:
                # CSS/Properties/text-align
                # valid values: http://www.datypic.com/sc/ooxml/t-w_ST_Jc.html
                align = {
                    "start": u"left",
                    "end": u"right",
                    "left": u"left",
                    "right": u"right",
                    "center": u"center",
                    "both": u"justify",
                    "distribute": u"justify",
                    # "mediumKashida": None,
                    # "numTab": None,
                    # "lowKashida": None,
                    # "thaiDistribute": None
                }[w_jc]
                styles["align"] = align

            # -- Borders
            w_tc_borders = value_of(w_tc, 'w:tcPr/w:tcBorders')
            cell_borders = _get_table_borders(w_tc_borders)
            styles.update(cell_borders)

            # todo: calculate the ``@rotate`` attribute.

            content = w_tc.xpath('w:p | w:tbl', namespaces=NS)

            # ignore the *tail* (if the XML is indented)
            for node in content:
                node.tail = None

            # The detection of empty cells (without text or image) is used when converting
            # to the Formex4 format in order to insert an empty tag ``<IE/>``.
            # see: https://github.com/laurent-laporte-pro/benker/issues/13
            if (
                not get_content_text(content)
                and not w_tc.xpath("count(.//w:drawing)", namespaces=NS)
                and not w_tc.xpath("count(.//w:pict)", namespaces=NS)
            ):
                # The cell has no text or image.
                styles["x-cell-empty"] = "true"

            state.row.insert_cell(content, width=width, height=height, styles=styles)



# coding: utf-8
"""
CALS - Frame Styles
===================

.. versionadded:: 0.5.0
"""

#: Default value for a solid border (for @cals:frame/@cals:colsep/@cals:rowsep, ...)
BORDER_SOLID = "solid 1pt black"

#: Default value for no border (for @cals:frame/@cals:colsep/@cals:rowsep, ...)
BORDER_NONE = "none"


def get_frame_styles(frame):
    styles = {}
    if not frame:
        return styles
    top, bottom, left, right = {
        "none": (False, False, False, False),
        "all": (True, True, True, True),
        "topbot": (True, True, False, False),
        "sides": (False, False, True, True),
        "top": (True, False, False, False),
        "bottom": (False, True, False, False),
    }[frame]
    styles["border-top"] = BORDER_SOLID if top else BORDER_NONE
    styles["border-bottom"] = BORDER_SOLID if bottom else BORDER_NONE
    styles["border-left"] = BORDER_SOLID if left else BORDER_NONE
    styles["border-right"] = BORDER_SOLID if right else BORDER_NONE
    return styles



# coding: utf-8
"""
OOXML Standard Types
====================

Utility classes used to store/convert OOXML standard types.
"""


class StValue(object):
    """
    Base class of ``ST_*`` types.
    """

    def __init__(self, value, default=None):
        if value == default:
            value = None
        self.value = value

    def __repr__(self):
        cls = self.__class__.__name__
        return "<{cls}({value})>".format(cls=cls, value=self.value)

    def __str__(self):
        return "" if self.value is None else self.value

    @property
    def style(self):
        return self.value


class StTwipsMeasure(StValue):
    """
    Measurement in Twentieths of a Point
    """

    @property
    def style(self):
        return None if self.value is None else float(self.value) / 20


class StPageOrientation(StValue):
    """
    Page Orientation

    http://www.datypic.com/sc/ooxml/t-w_ST_PageOrientation.html
    """


class StHexColor(StValue):
    """
    Color Value
    """

    def __init__(self, value):
        super(StHexColor, self).__init__(value, default="auto")

    @property
    def style(self):
        return None if self.value is None else "#{}".format(self.value)



# coding: utf-8
import pytest

from benker.parsers.cals.frame_styles import BORDER_NONE
from benker.parsers.cals.frame_styles import BORDER_SOLID
from benker.parsers.cals.frame_styles import get_frame_styles


@pytest.mark.parametrize(
    "frame, expected",
    [
        (None, {}),
        (
            "none",
            {
                "border-bottom": BORDER_NONE,
                "border-left": BORDER_NONE,
                "border-right": BORDER_NONE,
                "border-top": BORDER_NONE,
            },
        ),
        (
            "all",
            {
                "border-bottom": BORDER_SOLID,
                "border-left": BORDER_SOLID,
                "border-right": BORDER_SOLID,
                "border-top": BORDER_SOLID,
            },
        ),
        (
            "topbot",
            {
                "border-bottom": BORDER_SOLID,
                "border-left": BORDER_NONE,
                "border-right": BORDER_NONE,
                "border-top": BORDER_SOLID,
            },
        ),
        (
            "sides",
            {
                "border-bottom": BORDER_NONE,
                "border-left": BORDER_SOLID,
                "border-right": BORDER_SOLID,
                "border-top": BORDER_NONE,
            },
        ),
        (
            "top",
            {
                "border-bottom": BORDER_NONE,
                "border-left": BORDER_NONE,
                "border-right": BORDER_NONE,
                "border-top": BORDER_SOLID,
            },
        ),
        (
            "bottom",
            {
                "border-bottom": BORDER_SOLID,
                "border-left": BORDER_NONE,
                "border-right": BORDER_NONE,
                "border-top": BORDER_NONE,
            },
        ),
    ],
)
def test_get_frame_styles(frame, expected):
    actual = get_frame_styles(frame)
    assert actual == expected



# coding: utf-8
import pytest
from lxml import etree

from benker.builders.base_builder import BaseBuilder
from benker.parsers.ooxml import OoxmlParser
from benker.table import Table

TEST_DATA = [
    # -------------
    # without style
    # -------------
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:p w:rsidR="00EF2ECA" w:rsidRDefault="00EF2ECA"><w:r><w:t>empty</w:t></w:r></w:p>
        </w:tc>""",
        {},
        id="no_style",
    ),
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr>
                <w:tcW w:w="736" w:type="dxa"/>
                <w:gridSpan w:val="2"/>
            </w:tcPr>
            <w:p w:rsidR="00EF2ECA" w:rsidRDefault="00EF2ECA"><w:r><w:t>B</w:t></w:r></w:p>
        </w:tc>""",
        {},
        id="style_is_empty",
    ),
    # -------
    # Borders
    # -------
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr>
                <w:tcW w:w="1298" w:type="dxa"/>
                <w:gridSpan w:val="3"/>
                <w:tcBorders>
                    <w:bottom w:val="nil"/>
                    <w:right w:val="single" w:sz="12" w:space="0" w:color="4472C4" w:themeColor="accent1"/>
                </w:tcBorders>
                <w:shd w:val="clear" w:color="auto" w:fill="FBE4D5" w:themeFill="accent2" w:themeFillTint="33"/>
            </w:tcPr>
            <w:p w:rsidR="00EF2ECA" w:rsidRDefault="00EF2ECA"><w:r><w:t>1</w:t></w:r></w:p>
        </w:tc>""",
        {
            "border-collapse": "collapse",
            "border-right": "solid 1.5pt #4472C4",
            "border-bottom": "none",
            "background-color": "#FBE4D5",
        },
        id="style_border-right_solid_border-bottom_none",
    ),
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr>
                <w:tcW w:w="562" w:type="dxa"/>
                <w:tcBorders>
                    <w:top    w:val="double" w:sz="8"  w:color="FFFFFF"/>
                    <w:right  w:val="dotted" w:sz="16" w:color="FF0000"/>
                    <w:bottom w:val="dashed" w:sz="24" w:color="00FF00"/>
                    <w:left   w:val="outset" w:sz="32" w:color="0000FF"/>
                </w:tcBorders>
            </w:tcPr>
            <w:p w:rsidR="00EF2ECA" w:rsidRDefault="00EF2ECA"/>
        </w:tc>""",
        {
            "border-top": "double 1.0pt #FFFFFF",
            "border-right": "dotted 2.0pt #FF0000",
            "border-bottom": "dashed 3.0pt #00FF00",
            "border-left": "outset 4.0pt #0000FF",
        },
        id="style_border-all",
    ),
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr>
                <w:tcW w:w="562" w:type="dxa"/>
                <w:tcBorders>
                    <w:tr2bl w:val="single" w:sz="4" w:space="0" w:color="auto"/>
                    <w:tl2br w:val="single" w:sz="4" w:space="0" w:color="auto"/>
                </w:tcBorders>
            </w:tcPr>
            <w:p w:rsidR="00EF2ECA" w:rsidRDefault="00EF2ECA"/>
        </w:tc>""",
        {
            "border-collapse": "collapse",
            "x-border-tl2br": "solid 0.5pt",
            "x-border-tr2bl": "solid 0.5pt",
        },
        id="style_border-tl2br-tr2bl",
    ),
    # ------------------------
    # align and vertical-align
    # ------------------------
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr><w:vAlign w:val="top"/></w:tcPr>
            <w:p>
                <w:pPr><w:jc w:val="left"/></w:pPr>
                <w:r><w:t>text</w:t></w:r>
            </w:p>
        </w:tc>""",
        {"vertical-align": "top", "align": "left"},
        id="style_vertical-align_top_align_left",
    ),
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr><w:vAlign w:val="bottom"/></w:tcPr>
            <w:p>
                <w:pPr><w:jc w:val="right"/></w:pPr>
                <w:r><w:t>text</w:t></w:r>
            </w:p>
        </w:tc>""",
        {"vertical-align": "bottom", "align": "right"},
        id="style_vertical-align_bottom_align_right",
    ),
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr><w:vAlign w:val="center"/></w:tcPr>
            <w:p>
                <w:pPr><w:jc w:val="center"/></w:pPr>
                <w:r><w:t>text</w:t></w:r>
            </w:p>
        </w:tc>""",
        {"vertical-align": "middle", "align": "center"},
        id="style_vertical-align_middle_align_center",
    ),
    pytest.param(
        u"""<w:tc xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:tcPr><w:vAlign w:val="both"/></w:tcPr>
            <w:p>
                <w:pPr><w:jc w:val="both"/></w:pPr>
                <w:r><w:t>text</w:t></w:r>
            </w:p>
        </w:tc>""",
        {"vertical-align": "w-both", "align": "justify"},
        id="style_vertical-align_both_align_justify",
    ),
]


@pytest.mark.parametrize("w_tc_content, expected", TEST_DATA)
def test_parse_tc(w_tc_content, expected):
    builder = BaseBuilder()
    parser = OoxmlParser(builder)

    # -- setup a minimal table
    state = parser._state
    state.table = Table()
    state.row = state.table.rows[1]
    state.next_col()

    # -- parse a <w:tc/>
    w_tc = etree.XML(w_tc_content)
    parser.parse_tc(w_tc)

    # -- check the styles
    table = state.table
    cell = table[(1, 1)]

    # Ignore cell styles extensions (like 'x-cell-empty').
    actual = {k: v for k, v in cell.styles.items() if not k.startswith("x-cell-")}
    assert expected == actual



# coding: utf-8

import textwrap

from benker.table import Table


def test_colspan():
    # fmt: off
    table = Table()
    table.cols[1].insert_cell("one")
    table.cols[1].insert_cell("spanned", width=2)
    table.cols[2].insert_cell("two")
    text = str(table)
    assert text == textwrap.dedent("""\
    +-----------+-----------+
    |    one    |    two    |
    +-----------------------+
    |  spanned              |
    +-----------------------+""")
    # fmt: on


def test_rowspan():
    # fmt: off
    table = Table()
    table.rows[1].insert_cell("one")
    table.rows[1].insert_cell("spanned", height=2)
    table.rows[2].insert_cell("two")
    text = str(table)
    assert text == textwrap.dedent("""\
    +-----------+-----------+
    |    one    |  spanned  |
    +-----------|           |
    |    two    |           |
    +-----------+-----------+""")
    # fmt: on


def test_fill_missing_cells():
    # fmt: off
    table = Table()
    table.rows[1].insert_cell("one")
    table.rows[1].insert_cell("two")
    table.rows[1].insert_cell("three")
    table.rows[1].insert_cell("four", height=2)
    table.rows[2].insert_cell("un-deux", width=2)

    table.fill_missing(table.bounding_box, "???")
    text = str(table)
    assert text == textwrap.dedent("""\
    +-----------+-----------+-----------+-----------+
    |    one    |    two    |   three   |   four    |
    +-----------------------+-----------|           |
    |  un-deux              |    ???    |           |
    +-----------------------+-----------+-----------+""")
    # fmt: on



# coding: utf-8
"""
OOXML Page size
===============

Page Size dimensions, orientation and printer paper code.
"""

from benker.parsers.ooxml.namespaces import value_of
from benker.parsers.ooxml.types import StPageOrientation
from benker.parsers.ooxml.types import StTwipsMeasure


class PgSz(object):
    """
    Page Size. --
    This element specifies the properties (size and orientation)
    for all pages in the current section.

    Example: ``<w:pgSz w:w="11907" w:h="16839" />``, for A4 portrait.
    """

    def __init__(self, w_pg_sz):
        #: Page Width
        self.w_w = value_of(w_pg_sz, "@w:w")

        #: Page Height
        self.w_h = value_of(w_pg_sz, "@w:h")

        #: Page Orientation (Possible values are "landscape" and "portrait").
        self.w_orient = value_of(w_pg_sz, "@w:orient")

        #: Printer Paper Code
        self.w_code = value_of(w_pg_sz, "@w:code")

    @property
    def styles(self):
        w = StTwipsMeasure(self.w_w).style
        h = StTwipsMeasure(self.w_h).style
        size = (w, h) if w is not None and h is not None else None
        orient = StPageOrientation(self.w_orient).style
        styles = {"x-sect-size": size, "x-sect-orient": orient}
        return {k: v for k, v in styles.items() if v is not None}



# coding: utf-8
import textwrap
import unittest

import pytest
from lxml import etree
from lxml.builder import ElementMaker

from benker.builders.base_builder import BaseBuilder
from benker.parsers.cals.frame_styles import BORDER_NONE
from benker.parsers.cals.frame_styles import BORDER_SOLID
from benker.parsers.formex import FormexParser
from benker.schemas import CALS_NS
from benker.schemas import CALS_PREFIX
from benker.schemas import FORMEX_NS
from benker.schemas import FORMEX_PREFIX


class TestFormexParser(unittest.TestCase):
    def setUp(self):
        self.builder = BaseBuilder()

    def test_builder_attached(self):
        parser = FormexParser(self.builder)
        assert parser.builder is self.builder

    def test_ns(self):
        parser = FormexParser(self.builder)
        assert parser.formex_ns is None
        assert parser.cals_ns is None
        parser = FormexParser(self.builder, formex_ns="http://opoce")
        assert parser.formex_ns == "http://opoce"
        assert parser.cals_ns is None
        parser = FormexParser(self.builder, cals_ns=CALS_NS)
        assert parser.formex_ns is None
        assert parser.cals_ns == CALS_NS


class StrBuilder(BaseBuilder):
    """
    Create a ``table`` element with the string representation of the table object.
    """

    def generate_table_tree(self, table):
        element = etree.Element("table")
        element.text = str(table)
        return element


def test_transform_tables__no_namespace():
    E = ElementMaker()
    fmx_tbl = E.TBL(E.CORPUS(E.ROW(E.CELL("A1"), E.CELL("B1", ROWSPAN="2")), E.ROW(E.CELL("A2"))))
    tree = E.FORMEX(fmx_tbl)
    builder = StrBuilder()
    parser = FormexParser(builder)
    parser.transform_tables(tree)
    str_table = tree.xpath("//table")[0].text
    # print("str_table:")
    # print(str_table)
    # fmt: off
    assert str_table == textwrap.dedent("""\
    +-----------+-----------+
    |    A1     |    B1     |
    +-----------|           |
    |    A2     |           |
    +-----------+-----------+""")
    # fmt: on


def test_transform_tables__with_header():
    E = ElementMaker()
    # fmt: off
    fmx_tbl = E.TBL(E.CORPUS(
        E.ROW(E.CELL("Header 1", TYPE="HEADER"), E.CELL("Header 2", TYPE="HEADER"), TYPE="HEADER"),
        E.ROW(E.CELL("A1"), E.CELL("B1")),
        E.ROW(E.CELL("A2"), E.CELL("B2")),
        E.ROW(E.CELL("A3"), E.CELL("B3")),
    ))
    # fmt: on
    tree = E.FORMEX(fmx_tbl)
    builder = StrBuilder()
    parser = FormexParser(builder)
    parser.transform_tables(tree)
    str_table = tree.xpath("//table")[0].text
    # print("str_table:")
    # print(str_table)
    # fmt: off
    assert str_table == textwrap.dedent("""\
    +-----------+-----------+
    | Header 1  | Header 2  |
    +-----------+-----------+
    |    A1     |    B1     |
    +-----------+-----------+
    |    A2     |    B2     |
    +-----------+-----------+
    |    A3     |    B3     |
    +-----------+-----------+""")
    # fmt: on


def test_transform_tables__with_namespace():
    E = ElementMaker(namespace=FORMEX_NS, nsmap={FORMEX_PREFIX: FORMEX_NS})
    colspan = etree.QName(FORMEX_NS, "COLSPAN")
    # fmt: off
    tree = E.FORMEX(E.TBL(E.CORPUS(
        E.ROW(E.CELL("A1"), E.CELL("B1")),
        E.ROW(E.CELL("A2", **{colspan.text: "2"}))
    )))
    # fmt: on
    builder = StrBuilder()
    parser = FormexParser(builder, formex_prefix=FORMEX_PREFIX, formex_ns=FORMEX_NS)
    parser.transform_tables(tree)
    str_table = tree.xpath("//table")[0].text
    # print("str_table:")
    # print(str_table)
    # fmt: off
    assert str_table == textwrap.dedent("""\
    +-----------+-----------+
    |    A1     |    B1     |
    +-----------------------+
    |    A2                 |
    +-----------------------+""")
    # fmt: on


@pytest.mark.parametrize(
    "attrib, styles, nature",
    # fmt: off
    [
        (
            {"COLS": "1"},
            {},
            None,
        ),
        (
            {"NO.SEQ": "0001", "COLS": "1", "PAGE.SIZE": "DOUBLE.LANDSCAPE"},
            {"x-sect-orient": "landscape"},
            None,
        ),
        (
            {"NO.SEQ": "0001", "CLASS": "GEN", "COLS": "1", "PAGE.SIZE": "SINGLE.PORTRAIT"},
            {"x-sect-orient": "portrait"},
            "GEN",
        ),
        (
            {"NO.SEQ": "0001", "CLASS": "SCHEDULE", "COLS": "1", "PAGE.SIZE": "SINGLE.LANDSCAPE"},
            {"x-sect-orient": "landscape"},
            "SCHEDULE",
        ),
        (
            {"NO.SEQ": "0001", "CLASS": "RECAP", "COLS": "1", "PAGE.SIZE": "DOUBLE.PORTRAIT"},
            {"x-sect-orient": "portrait"},
            "RECAP",
        ),
    ],
    # fmt: on
)
def test_parse_fmx_tbl_corpus(attrib, styles, nature):
    fmx_tbl = etree.Element("TBL", attrib=attrib)
    fmx_corpus = etree.SubElement(fmx_tbl, "CORPUS")
    parser = FormexParser(BaseBuilder())
    state = parser.parse_fmx_corpus(fmx_corpus)
    table = state.table
    assert table.styles == styles
    assert table.nature == nature


@pytest.mark.parametrize(
    "attrib, styles",
    # fmt: off
    [
        (
            {"frame": "all"},
            {
                "border-bottom": BORDER_SOLID,
                "border-left": BORDER_SOLID,
                "border-right": BORDER_SOLID,
                "border-top": BORDER_SOLID,
            },
        ),
        (
            {"colsep": "0"},
            {"x-cell-border-right": BORDER_NONE},
        ),
        (
            {"colsep": "1"},
            {"x-cell-border-right": BORDER_SOLID},
        ),
        (
            {"rowsep": "0"},
            {"x-cell-border-bottom": BORDER_NONE},
        ),
        (
            {"rowsep": "1"},
            {"x-cell-border-bottom": BORDER_SOLID},
        ),
        (
            {"orient": "land"},
            {"x-sect-orient": "landscape"},
        ),
        (
            {"orient": "port"},
            {"x-sect-orient": "portrait"},
        ),
        (
            {"pgwide": "1"},
            {"x-sect-cols": "1"},
        ),
        (
            {"pgwide": "0"},
            {"x-sect-cols": "2"},
        ),
        (
            {"bgcolor": "#FF8000"},
            {"background-color": "#FF8000"},
        ),
    ]
    # fmt: on
)
def test_parse_fmx_corpus(attrib, styles):
    fmx_corpus = etree.Element("CORPUS", attrib=attrib)
    parser = FormexParser(BaseBuilder())
    state = parser.parse_fmx_corpus(fmx_corpus)
    table = state.table
    assert table.styles == styles


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({}, {}, "body"),
        ({"TYPE": "ALIAS"}, {"rowstyle": "ROW-ALIAS"}, "header"),
        ({"TYPE": "HEADER"}, {"rowstyle": "ROW-HEADER"}, "header"),
        ({"TYPE": "NORMAL"}, {"rowstyle": "ROW-NORMAL"}, "body"),
        ({"TYPE": "NOTCOL"}, {"rowstyle": "ROW-NOTCOL"}, "body"),
        ({"TYPE": "TOTAL"}, {"rowstyle": "ROW-TOTAL"}, "body"),
        ({"TYPE": "NORMAL", "rowstyle": "ROW-TOTAL"}, {"rowstyle": "ROW-TOTAL"}, "body"),
        ({"valign": "top"}, {"vertical-align": "top"}, "body"),
        ({"valign": "middle"}, {"vertical-align": "middle"}, "body"),
        ({"valign": "bottom"}, {"vertical-align": "bottom"}, "body"),
        ({"rowsep": "0"}, {"border-bottom": BORDER_NONE}, "body"),
        ({"rowsep": "1"}, {"border-bottom": BORDER_SOLID}, "body"),
        ({"bgcolor": "#FF8000"}, {"background-color": "#FF8000"}, "body"),
        ({"rowstyle": "RowStyle"}, {"rowstyle": "RowStyle"}, "body"),
        ({"TYPE": "ALIAS", "rowstyle": "RowStyle"}, {"rowstyle": "RowStyle"}, "header"),
    ],
)
def test_parse_fmx_row(attrib, styles, nature):
    E = ElementMaker()
    fmx_row = E.ROW(**attrib)
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state = parser.parse_fmx_row(fmx_row)
    row = state.row
    assert row.styles == styles
    assert row.nature == nature


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({}, {"rowstyle": "ROW-level2"}, "body"),
        ({"TYPE": "ALIAS"}, {"rowstyle": "ROW-ALIAS-level2"}, "header"),
        ({"TYPE": "HEADER"}, {"rowstyle": "ROW-HEADER-level2"}, "header"),
        ({"TYPE": "NORMAL"}, {"rowstyle": "ROW-NORMAL-level2"}, "body"),
        ({"TYPE": "NOTCOL"}, {"rowstyle": "ROW-NOTCOL-level2"}, "body"),
        ({"TYPE": "TOTAL"}, {"rowstyle": "ROW-TOTAL-level2"}, "body"),
    ],
)
def test_parse_fmx_row__in_blk_level2(attrib, styles, nature):
    E = ElementMaker()
    fmx_row = E.ROW(**attrib)
    E.BLK(E.BLK(fmx_row))
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state = parser.parse_fmx_row(fmx_row)
    row = state.row
    assert row.styles == styles
    assert row.nature == nature


def test_parse_fmx_ti_blk__level1():
    fmx_blk = etree.Element("BLK")
    fmx_ti_blk = etree.XML(
        """<TI.BLK COL.START="1" COL.END="2" valign="top" rowsep="1" bgcolor="blue"><P>paragraph</P></TI.BLK>"""
    )
    fmx_blk.append(fmx_ti_blk)
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state = parser.parse_fmx_ti_blk(fmx_ti_blk)
    row = state.row
    assert row.styles == {
        'background-color': 'blue',
        'border-bottom': 'solid 1pt black',
        'rowstyle': 'TI.BLK-level1',
        'vertical-align': 'top',
    }
    assert row.nature == "body"
    table = state.table
    cell = table[(1, 1)]
    assert cell.styles == {}
    assert cell.nature == "body"
    assert cell.width == 2
    assert cell.height == 1
    assert etree.tounicode(cell.content[0]) == "<P>paragraph</P>"


def test_parse_fmx_ti_blk__level2():
    fmx_blk = etree.Element("BLK")
    fmx_blk = etree.SubElement(fmx_blk, "BLK")
    fmx_ti_blk = etree.XML("""<TI.BLK><IE/></TI.BLK>""")
    fmx_blk.append(fmx_ti_blk)
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state = parser.parse_fmx_ti_blk(fmx_ti_blk)
    row = state.row
    assert row.styles == {"rowstyle": "TI.BLK-level2"}
    assert row.nature == "body"
    table = state.table
    cell = table[(1, 1)]
    assert cell.styles == {}
    assert cell.nature == "body"
    assert cell.width == 1
    assert cell.height == 1
    assert cell.content == ""


def test_parse_fmx_ti_blk__level2__with_namespace():
    def fmx(name):
        return etree.QName(FORMEX_NS, name).text

    BLK = fmx("BLK")
    TI_BLK = fmx("TI.BLK")
    IE = fmx("IE")

    fmx_blk1 = etree.Element(BLK, nsmap={None: FORMEX_NS})
    fmx_blk2 = etree.SubElement(fmx_blk1, BLK, nsmap={None: FORMEX_NS})
    fmx_ti_blk = etree.SubElement(fmx_blk2, TI_BLK, nsmap={None: FORMEX_NS})
    etree.SubElement(fmx_ti_blk, IE, nsmap={None: FORMEX_NS})

    parser = FormexParser(BaseBuilder(), formex_ns=FORMEX_NS, cals_prefix=CALS_PREFIX, cals_ns=CALS_NS)
    state = parser.setup_table()
    state.next_row()
    state = parser.parse_fmx_ti_blk(fmx_ti_blk)

    row = state.row
    assert row.styles == {"rowstyle": "TI.BLK-level2"}
    assert row.nature == "body"
    table = state.table
    cell = table[(1, 1)]
    assert cell.styles == {}
    assert cell.nature == "body"
    assert cell.width == 1
    assert cell.height == 1
    assert cell.content == ""


def test_parse_fmx_sti_blk__level1():
    fmx_blk = etree.Element("BLK")
    fmx_sti_blk = etree.XML(
        """<STI.BLK COL.START="2" COL.END="2" valign="top" rowsep="1" bgcolor="blue">text</STI.BLK>"""
    )
    fmx_blk.append(fmx_sti_blk)
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state = parser.parse_fmx_sti_blk(fmx_sti_blk)
    row = state.row
    assert row.styles == {
        'background-color': 'blue',
        'border-bottom': 'solid 1pt black',
        'rowstyle': 'STI.BLK-level1',
        'vertical-align': 'top',
    }
    assert row.nature == "body"
    table = state.table
    cell1 = table[(1, 1)]
    assert cell1.content is None
    cell2 = table[(2, 1)]
    assert cell2.styles == {}
    assert cell2.nature == "body"
    assert cell2.width == 1
    assert cell2.height == 1
    assert cell2.content[0] == "text"


def test_parse_fmx_gr_notes():
    # fmt: off
    fmx_gr_notes = etree.XML("""\
    <GR.NOTES valign="top" rowsep="1" bgcolor="blue">
      <TITLE><TI><P>GR.NOTES Title</P></TI></TITLE>
      <NOTE NOTE.ID="N0001"><P>Table note</P></NOTE>
    </GR.NOTES>""")
    # fmt: on

    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    # -- insert at least one ROW for testing
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state.row.insert_cell("text1")
    state.row.insert_cell("text2")
    state.row.insert_cell("text3")
    # -- then add the footer
    state.next_row()
    state = parser.parse_gr_notes(fmx_gr_notes)
    row = state.row
    assert row.styles == {}
    assert row.nature == "footer"
    # -- the cell is in the row 2
    cell = state.table[(1, 2)]
    assert cell.styles == {'background-color': 'blue', 'border-bottom': 'solid 1pt black', 'vertical-align': 'top'}
    assert cell.nature == "footer"
    assert cell.width == 3
    assert cell.height == 1
    content = [node for node in cell.content if isinstance(node, etree._Element)]
    assert etree.tounicode(content[0], with_tail=False) == "<TITLE><TI><P>GR.NOTES Title</P></TI></TITLE>"
    assert etree.tounicode(content[1], with_tail=False) == '<NOTE NOTE.ID="N0001"><P>Table note</P></NOTE>'


def test_parse_fmx_gr_notes__embed_gr_notes():
    # fmt: off
    gr_notes = """\
    <GR.NOTES valign="top" rowsep="1" bgcolor="blue">
      <TITLE><TI><P>GR.NOTES Title</P></TI></TITLE>
      <NOTE NOTE.ID="N0001"><P>Table note</P></NOTE>
    </GR.NOTES>"""
    fmx_gr_notes = etree.XML(gr_notes)
    # fmt: on

    parser = FormexParser(BaseBuilder(), embed_gr_notes=True)
    state = parser.setup_table()
    # -- insert at least one ROW for testing
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    state.row.insert_cell("text1")
    state.row.insert_cell("text2")
    state.row.insert_cell("text3")
    # -- then add the footer
    state.next_row()
    state = parser.parse_gr_notes(fmx_gr_notes)
    row = state.row
    assert row.styles == {}
    assert row.nature == "footer"
    # -- the cell is in the row 2
    cell = state.table[(1, 2)]
    assert cell.styles == {'background-color': 'blue', 'border-bottom': 'solid 1pt black', 'vertical-align': 'top'}
    assert cell.nature == "footer"
    assert cell.width == 3
    assert cell.height == 1
    content = [node for node in cell.content if isinstance(node, etree._Element)]
    assert etree.tounicode(content[0]) == etree.tounicode(fmx_gr_notes)


@pytest.mark.parametrize(
    "attrib, styles, nature, size",
    [
        ({}, {}, None, (1, 1)),
        ({"COLSPAN": "1"}, {}, None, (1, 1)),
        ({"COLSPAN": "2"}, {}, None, (2, 1)),
        ({"ROWSPAN": "1"}, {}, None, (1, 1)),
        ({"ROWSPAN": "2"}, {}, None, (1, 2)),
        ({"TYPE": "ALIAS"}, {"cellstyle": "ALIAS"}, "header", (1, 1)),
        ({"TYPE": "HEADER"}, {"cellstyle": "HEADER"}, "header", (1, 1)),
        ({"TYPE": "NORMAL"}, {}, "body", (1, 1)),
        ({"TYPE": "NOTCOL"}, {"cellstyle": "NOTCOL"}, "body", (1, 1)),
        ({"TYPE": "TOTAL"}, {"cellstyle": "TOTAL"}, "body", (1, 1)),
    ],
)
def test_parse_fmx_cell(attrib, styles, nature, size):
    E = ElementMaker()
    fmx_cell = E.CELL(**attrib)
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    parser.parse_fmx_cell(fmx_cell)
    table = state.table
    cell = table[(1, 1)]
    assert cell.styles == styles
    assert cell.nature == nature
    assert cell.size == size


def test_parse_fmx_cell__with_cals():
    E = ElementMaker()
    fmx_cell = E.CELL(
        colsep="1",
        rowsep="1",
        namest="c1",
        nameend="c3",
        bgcolor="#00007f",
        morerows="1",
        align="center",
        valign="middle",
    )
    parser = FormexParser(BaseBuilder())
    state = parser.setup_table()
    state.next_row()
    state.row = state.table.rows[state.row_pos]
    parser.parse_fmx_cell(fmx_cell)
    table = state.table
    cell = table[(1, 1)]
    assert cell.styles == {
        'align': 'center',
        'background-color': '#00007f',
        'vertical-align': 'middle',
        'x-cell-border-bottom': 'solid 1pt black',
        'x-cell-border-right': 'solid 1pt black',
    }
    assert cell.nature is None
    assert cell.size == (3, 2)


@pytest.mark.parametrize(
    "attrib, styles, nature",
    [
        ({"colnum": "2"}, {}, None),
        ({"colsep": "0"}, {"border-right": BORDER_NONE}, None),
        ({"colsep": "1"}, {"border-right": BORDER_SOLID}, None),
        ({"rowsep": "0"}, {"border-bottom": BORDER_NONE}, None),
        ({"rowsep": "1"}, {"border-bottom": BORDER_SOLID}, None),
        ({"colwidth": "5mm"}, {"width": "5mm"}, None),
        ({"align": "left"}, {"align": "left"}, None),
        ({"align": "right"}, {"align": "right"}, None),
        ({"align": "center"}, {"align": "center"}, None),
        ({"align": "justify"}, {"align": "justify"}, None),
        ({"align": "char"}, {"align": "left"}, None),
    ],
)
def test_parse_fmx_colspec(attrib, styles, nature):
    cals_colspec = etree.Element("colspec", attrib=attrib)
    parser = FormexParser(BaseBuilder())
    parser.setup_table()
    state = parser._state
    state.next_col()
    state = parser.parse_fmx_colspec(cals_colspec)
    col_pos = int(attrib.get("colnum", 1))
    col = state.table.cols[col_pos]
    assert col.styles == styles
    assert col.nature == nature


@pytest.mark.parametrize(
    "content, expected",
    # fmt: off
    [
        ("<CELL>text</CELL>", False),
        ("<CELL><P>text</P></CELL>", False),
        ("<CELL><IE/></CELL>", True),
    ]
    # fmt: on
)
def test_contains_ie(content, expected):
    fmx_node = etree.XML(content)
    parser = FormexParser(BaseBuilder())
    assert parser.contains_ie(fmx_node) is expected


@pytest.mark.parametrize(
    "content, expected",
    [
        ('<CELL xmlns="http://opoce">text</CELL>', False),
        ('<CELL xmlns="http://opoce"><P>text</P></CELL>', False),
        ('<CELL xmlns="http://opoce"><IE/></CELL>', True),
    ],
)
def test_contains_ie__with_formex_ns(content, expected):
    fmx_node = etree.XML(content)
    parser = FormexParser(BaseBuilder(), formex_ns="http://opoce")
    assert parser.contains_ie(fmx_node) is expected



"""Easily convert your CALS, HTML, Formex 4, Office Open XML (docx) tables from one format to another."""

__version__ = "0.5.3"
__author__ = "Laurent LAPORTE <laurent.laporte.pro@gmail.com>"
__date__ = "2021-11-11"
__credit__ = "(c) 2018-2021 Laurent LAPORTE"



# coding: utf-8
"""
OOXML namespaces
================
"""
import functools

from benker.parsers.base_parser import value_of as base_value_of

#: Namespace map used for xpath evaluation in Office Open XML documents
NS = {'w': "http://schemas.openxmlformats.org/wordprocessingml/2006/main"}


def ns_name(ns, name):
    return '{' + ns + '}' + name


w = functools.partial(ns_name, NS['w'])
value_of = functools.partial(base_value_of, namespaces=NS)



# coding: utf-8
"""
lxml Iterators
==============

Python alternative to :class:`lxml.etree.iterwalk` for lxml < 4.2.1
"""
from lxml import etree

# noinspection PyProtectedMember
#: ElementTree Type
ElementTreeType = etree._ElementTree

# noinspection PyProtectedMember
#: Element Type
ElementType = etree._Element

if etree.LXML_VERSION >= (4, 2, 1, 0):
    iterwalk = etree.iterwalk

else:
    class _MultiTagMatcher(object):
        def __init__(self, tag):
            self.tags = {tag} if isinstance(tag, str) else set(tag)

        def __call__(self, element_or_tree):
            # type: (ElementTreeType or ElementType) -> bool
            return element_or_tree.tag in self.tags

    def _iterwalk_impl(tree, events, matcher):
        if 'start' in events and (matcher is None or matcher(tree)):
            yield ('start', tree)
        for child in tree:
            for event, node in _iterwalk_impl(child, events, matcher):
                yield (event, node)
        if 'end' in events and (matcher is None or matcher(tree)):
            yield ('end', tree)

    def iterwalk(element_or_tree, events=(u"end",), tag=None):
        matcher = None if tag is None or tag == "*" else _MultiTagMatcher(tag)
        for event, node in _iterwalk_impl(element_or_tree, events, matcher):
            yield (event, node)



# coding: utf-8
"""
.. _available-builders:

Available Builders
==================

This package contains a collection of builders which you can use in conjonction
with parsers to convert tables from one format to another.

You can pick a parser in the :ref:`Available Parsers <available-parsers>`.
"""
